from typing import Hashable


try:
    #! the wse allows import errors to pass silently!
    #* put all imports in this try block so that we can see the error before exiting
    
    import datetime
    import time
    import pdb
    from typing import Type

    from np_workflows.services import open_ephys as OpenEphys
    from np_workflows.services import middleware_messages_pb2 as router_msg

    from np_workflows.workflows.shared.npxc import experiment
    from np_workflows.workflows.shared import npxc
    from np_workflows.workflows.shared.initialize import (
        select_mouse_and_operator_enter, 
        select_mouse_and_operator_input,
        select_experiment_enter, 
        select_experiment_input,
        )
    from np_workflows.workflows.shared.photodoc import (
        capture_photodoc_input,
        review_photodoc_enter,
    )
    from np_workflows.workflows.shared.pretest import (
        run_pretest_input,
    )
    
except Exception as exc:
    print(repr(exc))
    import pdb; pdb.set_trace()
    exit()

npxc.start_rsc_apps()

# name each photodoc state `capture_photodoc_<state>`, `review_photodoc_<state>`
photodoc_states_to_labels = {
    0: 'brain_surface',
    1: 'pre_recording', # once per trial loop, after probes advanced + settled 
    2: 'post_experiment', # once, after all experiment loops
    }

# def generate_photodoc_states(photodoc_states_to_labels: dict[Hashable, str]):
#* can we put this in shared.photodoc module (/in a function) and import them automatically?
for state, label in photodoc_states_to_labels.items():
    exec(
        f"""def capture_photodoc_{state}_enter(state_globals):
            npxc.set(state_globals, photodoc_label='{label}')
            npxc.set(state_globals, msg_text='{label} will be captured')
        """
    )
    exec(
        f"""def capture_photodoc_{state}_input(state_globals): 
            capture_photodoc_input(state_globals)
        """
    )
    exec(
        f"""def review_photodoc_{state}_enter(state_globals):
            review_photodoc_enter(state_globals)
        """
    )

def start_experiment_loop_enter(state_globals) -> None:
    npxc.experiment.advance_trial_idx(state_globals)

def start_recording_input(state_globals) -> None:
    npxc.experiment.start_recording()
    
def start_stimulus_input(state_globals) -> None:
    npxc.experiment.start_stimulus()
        
def stop_recording_input(state_globals) -> None:
    npxc.experiment.stop_recording()
       
def settle_timer_enter(state_globals):
    
    import pdb; pdb.set_trace()
    if npxc.get(state_globals, 'override_settle_timer'):
        npxc.set(state_globals, await_timer=False)
        npxc.set(state_globals, override_settle_timer=True)
        return
    
    # TODO move time to experiment.config
    settle_timer_sec = 10
    
    if time_remaining := npxc.await_timer(state_globals, wait_sec=settle_timer_sec):
        time.sleep(.1) # when time is printed on-screen it will be higher than actual time remaining by at least this much
        npxc.set(state_globals, time_remaining=f'{time_remaining}')
        npxc.set(state_globals, override_settle_timer=False)
        npxc.set(state_globals, await_timer=True)
        state_globals['resources']['io'].write(router_msg.state_busy(message="Awaiting settle timer"))
        return
    # print('Finished waiting')
    # # import pdb; pdb.set_trace()
    npxc.set(state_globals, override_settle_timer=True)
    npxc.set(state_globals, await_timer=False)
    state_globals['resources']['io'].write(router_msg.state_ready(message="Finished awaiting settle timer"))

# def settle_timer_input(state_globals): 
#     npxc.set(state_globals, next_state='settle_timer')
#     # import pdb; pdb.set_trace()
    # if npxc.get(state_globals, 'override_settle_timer') or not npxc.get(state_globals, 'await_timer'):
    #     return
    # settle_timer_sec = 10
    # if time_remaining := npxc.await_timer(state_globals, wait_sec=settle_timer_sec):
    #     time.sleep(.1) # when time is printed on-screen will be higher than actual value by at least this much
    #     npxc.set(state_globals, time_remaining=f'{time_remaining}')
    #     npxc.set(state_globals, override_settle_timer=False)
    #     npxc.set(state_globals, await_timer=True)
    #     state_globals['resources']['io'].write(router_msg.state_busy(message="Awaiting settle timer"))
    #     return
    # state_globals['resources']['io'].write(router_msg.state_ready(message="Finished
    # awaiting settle timer"))
    

def settle_timer_enter(state_globals):
    state_globals['resources']['io'].write(router_msg.state_busy(message="Awaiting settle timer"))
    npxc.set(state_globals, override_settle_timer=False)
    # pass
# def settle_timer_input(state_globals):
    # npxc.set(state_globals, msg_text='Awaiting settle timer')
    import threading
    def wait_on_timer():
        settle_timer_sec = 10
        while time_remaining := npxc.await_timer(state_globals, wait_sec=settle_timer_sec):
            time.sleep(.1)
            if npxc.get(state_globals, 'override_settle_timer'):
                break
            npxc.set(state_globals, time_remaining=f'{time_remaining}')
        state_globals['resources']['io'].write(router_msg.state_ready(message="Finished awaiting settle timer"))
        # npxc.set(state_globals, msg_text='Finished awaiting settle timer')
        # npxc.set(state_globals, alert=True)
            
    t = threading.Thread(target=wait_on_timer)
    t.start()

def settle_timer_input(state_globals):
    if npxc.get(state_globals, 'override_settle_timer'):
        return
    npxc.set(state_globals, next_state='settle_timer')


import os
import platform
import re
from enum import Enum
import socket
from typing import List, Union

import requests
import np_config
import np_logging 

logger = np_logging.getLogger(__name__)

server = "http://mpe-computers/v2.0"
ALL_RIGS = requests.get(server).json()
ALL_COMPS = requests.get(server+"/aibs_comp_id").json()

# get AIBS IDs, if set
COMP_ID: str = os.environ.get("AIBS_COMP_ID", socket.gethostname()).upper()
RIG_ID: str = os.environ.get("AIBS_RIG_ID", None)

while not RIG_ID:
    
    # extract RIG_ID from COMP_ID if possible
    if "NP." in COMP_ID:
        str_match = re.search(R"NP.[\d]+", COMP_ID)
        if str_match:
            RIG_ID = str_match[0]
            break
    
    # use BTVTest.1 if allowed
    # set with environ var:
    USE_TEST_RIG = os.environ.get("USE_TEST_RIG", True)
    if USE_TEST_RIG:
        RIG_ID = "BTVTest.1"
        break
    
    RIG_ID = "none"
    print("Not running from an NP rig: connections to services won't be made\nTry setting env var USE_TEST_RIG=1")

print(f"Running from {COMP_ID}, connected to {RIG_ID}")


CONFIG = np_config.from_zk(f"/rigs/{RIG_ID}")

# class Rig(Enum):
#     SYNC = f"{RIG_ID}-Sync"
#     WSE = f"{RIG_ID}-Sync"
#     MON = f"{RIG_ID}-Mon"
#     VIDMON = f"{RIG_ID}-Mon"
#     STIM = f"{RIG_ID}-Stim"
#     CAMSTIM = f"{RIG_ID}-Stim"
#     ACQ = f"{RIG_ID}-Acq" # TODO add btvtest.1-Acq http://mpe-computers/
#     EPHYS = f"{RIG_ID}-Acq"
       
class Rig(Enum):
    
    wse = wse2 = "Sync"
    sync = Sync = SYNC = "Sync"
    mvr = Mvr = MVR = "Mon"
    mon = Mon = MON = vidmon = Vidmon = VIDMON = "Mon"
    cam3d = Cam3d = CAM3D = "Mon"
    camviewer = CamViewer = camViewer = CAMVIEWER = "Mon"
    mousedirector = Mousedirector = mouseDirector = MouseDirector = MOUSEDIRECTOR = "Mon"
    camstim = Camstim = CamStim = CAMSTIM = "Stim"
    stim = Stim = STIM = "Stim"
    acq = Acq = ACQ = "Acq" # TODO add btvtest.1-Acq http://mpe-computers/
    ephys = Ephys = EPhys = EPHYS = "Acq"
    oephys = Oephys = oEphys = OEphys = OEPHYS = "Acq"
    openephys = openEphys = OpenEphys = OPENEPHYS = "Acq"
    
    def __new__(cls,suffix):
        ID = None
        while not RIG_ID:
            
            # extract RIG_ID from COMP_ID if possible
            ID = cls.rig_str_with_digit(COMP_ID)
            if ID:
                break
            
            # use BTVTest.1 if allowed
            # set with environ var:
            USE_TEST_RIG = os.environ.get("USE_TEST_RIG", True)
            if USE_TEST_RIG:
                ID = "BTVTest.1"
                break
            
            ID = "none"
        if not (RIG_ID or ID):
            print("Not running from an NP rig: connections to services won't be made\nTry setting env var USE_TEST_RIG=1")
        cls.ID = RIG_ID or ID
        obj = object.__new__(cls)
        obj._value_ = cls.ID + "-" + suffix
        obj.AIBS_ID = f"{cls.ID}-{suffix}"
        return obj
        # cls(self.name).value = cls.ID + suffix
    
    @classmethod    
    @property
    def idx(cls) -> int:
        return int(cls.rig_str_to_int(RIG_ID))
    
    @property
    def host(self):
        if "BTVTest.1-Acq" == self.value: 
            # not in mpe-computers
            return ""
        try:
            return ALL_COMPS.get(self.value, {}).get('hostname','').upper()                      
        except KeyError:
            return 

    @property
    def path(self):
        if "BTVTest.1-Acq" == self.value: 
            # not in mpe-computers
            return None

        if platform.system() == "Windows":
            return RF'//{self.host}'
        else:
            return RF'/{self.host}'

    def open(self):
        return os.startfile(self.path)
    
    
    @staticmethod
    def comp():
        """The comp where python is running, regardless of whether on a rig"""
        return COMP_ID
    
    
    @staticmethod
    def comps():
        """All computers on the current rig (if python is running on a rig comp)"""
        return {Rig(x).value: Rig(x).host for x in set(Rig.__members__.values())}
    
    
    @staticmethod
    def all_comps(rigs=[0,1,2]):
        """All computers on all np rigs, or those specified"""
        return ConfigHTTP.get_np_computers(rigs)
    
    @staticmethod
    def rig_from_path(path):
        for idx, rig in enumerate(["NP.0", "NP.1", "NP.2"]):
            for comp in Rig.all_comps(idx).values():
                if comp in path:
                    return rig
        return None
        
    @staticmethod
    def rig_str_to_int(rig:str) -> Union[int,None]:
        # extract RIG_ID from str if possible
        str_match = re.search(R"(?<=NP\.)(\d)", rig.upper())
        if str_match:
            return str_match[0]
        return None
    
    @staticmethod
    def rig_str_with_digit(rig:str) -> Union[str,None]:
        # extract RIG_ID from str if possible
        str_match = re.search(R"NP.\d", rig.upper())
        if str_match:
            return str_match[0]
        return None
        
    
class ConfigHTTP:
    
    server = "http://mpe-computers/v2.0"
    # all_pc = requests.get("http://mpe-computers/v2.0").json()

    @staticmethod
    def hostname(comp: str=''):
        if "BTVTest.1-Acq" in comp: # not in mpe-computers
            return None
        else:
            return ALL_COMPS[comp]['hostname'].upper()                      
    
    @staticmethod
    def get_np_computers(rigs: Union[List[int], int]=None, comp: Union[List[str], str]=None):
        if rigs is None:
            rigs = [0, 1, 2]

        if not isinstance(rigs, list):
            rigs = [int(rigs)]

        if comp is None:
            comp = ['sync', 'acq', 'mon', 'stim']

        if not isinstance(comp, list):
            comp = [str(comp)]

        comp = [c.lower() for c in comp]

        np_idx = ["NP." + str(idx) for idx in rigs]

        a = {}
        for k, v in ALL_RIGS['comp_ids'].items():
            if any([sub in k for sub in np_idx]) and any([s in k.lower() for s in comp]):
                a[k] = v['hostname'].upper()
        return a



from __future__ import annotations

try:
    #! the wse allows import errors to pass silently!
    #* put all imports in this try block so that we can see the error before exiting
    
    import datetime
    import time
    from typing import Type

    from np_workflows.services.protocols import Startable, Stoppable, Finalizable
    from np_workflows.experiments import baseclasses, classes
    from np_workflows.experiments.baseclasses import Experiment
    from np_workflows.workflows.shared import npxc

    # we need all experiment classes imported into this namespace
    from np_workflows.experiments.classes import *

    import yaml
    import np_session
    import np_logging

except Exception as exc:
    print(repr(exc))
    import pdb; pdb.set_trace()
    quit()

logger = np_logging.getLogger(__name__)

# def capture_photodoc_enter(state_globals) -> None:
#     if not npxc.get(state_globals, "photodoc_labels"):
#         npxc.set(state_globals, photodoc_labels=[""] + list(npxc.experiment.photodoc_labels))
#     if (selected_label := npxc.get(state_globals, "photodoc_label")) and selected_label not in (dropdown_labels := npxc.get(state_globals, "photodoc_labels")):
#         npxc.set(state_globals, photodoc_labels=[selected_label] + dropdown_labels)
        
def capture_photodoc_input(state_globals) -> None:
    current_label = npxc.get(state_globals, "photodoc_label")
    for service in npxc.experiment.photodoc_services:
        service.label = current_label
        if isinstance(service, Startable):
            service.start()
        if isinstance(service, Stoppable):
            service.stop()
        if isinstance(service, Finalizable):
            service.finalize()
                
def review_photodoc_enter(state_globals) -> None:
    # default to re-taking the photodoc

    for service in npxc.experiment.photodoc_services:
        img = None
        with contextlib.suppress(AttributeError, IndexError):
            img = service.data_files[-1]
        if not img:
            continue
        if img.suffix in ('.jpg', '.jpeg', '.bmp', '.png'):
            npxc.set(state_globals, current_image=str(img))
            break
    else:
        import pdb; pdb.set_trace()
        raise ValueError(f"No photodoc image found from services {[_.__name__ for _ in npxc.experiment.photodoc_services]}")
    


try:
    import inspect
    import json
    import logging
    import os
    import socket
    import time
    import webbrowser

    # from physync.report import NeuropixelsReport
        # -*- coding: latin-1 -*-
    from datetime import date as date
    from datetime import datetime as dt
    from datetime import timedelta as timedelta
    from pprint import pformat

    import np.workflows.npxcommon as npxc
    import requests
    from mpetk import limstk
    from mpetk.aibsmw.routerio.router import ZMQHandler
    from mpetk.zro import Proxy
    from np.models.model import Passive, VariabilitySpontaneous
except Exception as e:
    # import errors aren't printed to console by default
    print(e)

# -------------- experiment-specific objects --------------
global experiment
# this should be the first line of code executed
experiment = npxc.experiment = VariabilitySpontaneous()

global config
config = npxc.config = npxc.get_config()

# ---------------- Network Service Objects ----------------

router: ZMQHandler = None
camstim: Proxy = None
mouse_director: Proxy = None
sync: Proxy = None



# @npxc.state_transition
def default_exit(state_globals,label):
    npxc.default_exit(state_globals,label)

# @npxc.state_transition
def default_enter(state_globals,label):
    npxc.default_enter(state_globals,label)

# @npxc.state_transition
def default_input(state_globals, label):
    npxc.default_input(state_globals, label)


@npxc.state_transition
def overrideable_error_state_input(state_globals):
    npxc.overrideable_error_state_input(state_globals)

@npxc.state_transition
def overrideable_error_state_exit(state_globals):
    npxc.overrideable_error_state_exit(state_globals)

# @npxc.state_transition
def initialize_enter(state_globals):
    
    state_globals["external"]["logo"] = R".\np\images\logo_np_hab.png"

    state_globals['external']['session_type'] = 'behavior_habituation'
    npxc.initialize_enter(state_globals)


def initialize_input(state_globals):
    """
    Input test function for state initialize
    """
    npxc.initialize_input(state_globals)

    global mouse_director
    md_host = npxc.config['components']['MouseDirector']['host']
    md_port = npxc.config['components']['MouseDirector']['port']
    #! should md_host be localhost, not vidmon?
    md_host = 'localhost'
    mouse_director = Proxy(f'{md_host}:{md_port}')  # TODO: config this

    global camstim
    host = npxc.config["components"]["Stim"]["host"]
    port = npxc.config["components"]["Stim"]["port"]
    state_globals['mtrain_text'] =  pformat(npxc.config['MTrainData'])
    camstim = Proxy(f"{host}:{port}", serialization="json")

    global router
    router = state_globals["resources"]["io"]


    beh_mon_str_def = 'Limited adjustment to the wheel and the spout location should be made as necessary. '
    state_globals['external']['behavior_monitoring_string'] = npxc.get_from_config(['behavior_monitoring_string_hab'], default=beh_mon_str_def)


    defaults = ['first', 'normal', '2_days_before_experiment', '1_day_before_experiment']
    key_list = ['Habituaiton_Sessions']
    hab_sessions = npxc.get_from_config(key_list, defaults)
    #hab_sessions = []
    #try:
    #    hab_sessions = npxc.config['Habituaiton_Sessions']
    #except Exception as E:
    #    message = f'Failed to find habitaution sessions in  config. Using default instead'
    #    print(message)
    #    #npxc.alert_text(message, state_globals)
    #    hab_sessions = ['first', 'normal', '2_days_before_experiment', '1_day_before_expeirment']

    #session_list = [hab_session.lower() for hab_session in hab_sessions]
    state_globals["external"]["session_type_option_string"] = ', '.join(hab_sessions)
    state_globals["external"]["session_types_options"] = hab_sessions

    logging.info(
        f'Neuropixel workflow running on {state_globals["external"]["rig_id"]} '
        f'by user:{state_globals["external"]["user_id"]}',
        extra={"weblog": True},
    )

    result = limstk.user_details(state_globals["external"]["user_id"])
    failed = npxc.confirm_components(state_globals)

    state_globals["external"]["next_state"] = "scan_mouse_id"
    if failed:
        state_globals["external"]["next_state"] = "components_error"
        alert_string = f'The following proxies are not available: {", ".join(failed)}'
        npxc.alert_text(alert_string, state_globals)#Todo put this back  - I think it was being wierd.

    #else:
    #    state_globals["external"]["next_state"] = "components_error"
    try:
        if result != -1:
            state_globals["external"]["lims_user_id"] = result[0]["id"]
            state_globals["external"]["status_message"] = "success"
            state_globals["external"][
                "local_log"
            ] = f'LIMS ID for User:{state_globals["external"]["user_id"]} Found in LIMS'
        else:
            state_globals["external"]["next_state"] = "initialize"
            print('Failed user ID test')
            npxc.fail_state(f'No LIMS ID for User:{state_globals["external"]["user_id"]} found in LIMS', state_globals)
    except (KeyError, IndexError):
        print('Failed user ID test')
        npxc.fail_state(f'No LIMS ID for User:{state_globals["external"]["user_id"]} found in LIMS', state_globals)
        state_globals["external"]["next_state"] = "initialize"


    # state_globals['external']['next_state'] = 'probes_final_depth'  # NOTE: REMOVE this




def components_error_input(state_globals):
    npxc.components_error_input(state_globals, 'scan_mouse_id')


def mtrain_change_stage_enter(state):

    state["external"]["current_regimen"] = npxc.mtrain.regimen['name']
    state["external"]["current_stage"] = npxc.mtrain.stage['name'].title()
    state["external"]["new_stage"] = npxc.mtrain.stage['name'].title()

    available_stages = sorted([stage['name'].title() for stage in npxc.mtrain.stages])

    state["external"]["available_stages"] = npxc.circshift_to_item(available_stages, state['external']['current_stage'])

    # state['resources']['io'].write(npxc.messages.state_ready(message="ready"))


def mtrain_change_stage_input(state):

    print(state["external"]["next_state"])
    new_stage = state["external"]["new_stage"]
    confirm_stage = state['external']['confirm_stage']
    change_regimen = state['external']['change_regimen']

    if confirm_stage and new_stage.lower() != npxc.mtrain.stage['name'].lower():
        npxc.mtrain.stage = new_stage
        state["external"]["next_state"] = 'mtrain_change_stage'

    elif change_regimen:
        state["external"]["next_state"] = 'mtrain_change_regimen_1'


def mtrain_change_regimen_1_enter(state):
    # current regimen is already set in prev screen, but we set it again here anyway
    state["external"]["current_regimen"] = npxc.mtrain.regimen['name']

    available_regimens = sorted([regimen for regimen in npxc.mtrain.all_regimens().values()])
    state["external"]["available_regimens"] = npxc.circshift_to_item(available_regimens,
                                                                     state['external']['current_regimen'])


def mtrain_change_regimen_1_input(state):
    new_regimen = state["external"].get("new_regimen", None)
    confirm_regimen = state['external']['confirm_regimen']
    cancel_regimen_select = state['external']['cancel_regimen_select']

    if cancel_regimen_select:
        # cancel
        state["external"]["next_state"] = 'mtrain_change_stage'
    elif confirm_regimen and new_regimen:
        # don't set anything yet - we need a corresponding stage for the new regimen
        state["external"]["next_state"] = 'mtrain_change_regimen_2'
    elif confirm_regimen and not new_regimen:
        # apparently no change is requested - go back to the stage selection
        state["external"]["next_state"] = 'mtrain_change_stage'


def mtrain_change_regimen_2_enter(state):
    # get the stages available for the new regimen:
    new_regimen_dict = [
        regimen for regimen in npxc.mtrain.get_all("regimens")
        if regimen['name'].lower() == state["external"]["new_regimen"].lower()
    ][0]
    new_stages = new_regimen_dict['stages']
    # no need to circshift the list below, since there's no concept of "current stage" on a newly selected regimen
    state['external']['available_stages_new_regimen'] = sorted([stage['name'].title() for stage in new_stages])
    state["external"]["new_regimen"] = new_regimen_dict['name']


def mtrain_change_regimen_2_input(state):
    # copy verbatim this line from mtrain_change_regimen_2_enter
    new_regimen_dict = [
        regimen for regimen in npxc.mtrain.get_all("regimens")
        if regimen['name'].lower() == state["external"]["new_regimen"].lower()
    ][0]

    selected_stage_new_regimen = state["external"].get("selected_stage_new_regimen", None)
    confirm_regimen_and_stage = state['external']['confirm_regimen_and_stage']
    cancel_regimen_select = state['external']['cancel_regimen_select']

    # pdb.set_trace()
    # state['external']['confirm_stage_or_change_regimen']
    #TODO get the next state(s) from brb in wfl/state if possible

    if cancel_regimen_select:
        state["external"]["next_state"] = 'mtrain_change_stage'

    elif confirm_regimen_and_stage and selected_stage_new_regimen:

        new_stage_dict = [
            stage for stage in new_regimen_dict['stages']
            if stage['name'].lower() == selected_stage_new_regimen.lower()
        ][0]
        npxc.mtrain.set_regimen_and_stage(new_regimen_dict, new_stage_dict)
        state["external"]["next_state"] = 'mtrain_change_stage'

    elif confirm_regimen_and_stage and not selected_stage_new_regimen:
        # state["external"]["next_state"] = 'run_stimulus'
        # not sure if it's even possible to continue (green arrow) without selecting a stage
        state["external"]["next_state"] = 'mtrain_change_stage'



def pretest_input(state_globals):
    print('>> pretest_error_input <<')
    npxc.handle_2_choice_button('pretest_failed', 'pretest_error', 'configure_hardware', state_globals)


def pretest_error_input(state_globals):
    print('>> pretest_error_input <<')
    npxc.handle_2_choice_button('pretest_override', 'configure_hardware', 'pretest', state_globals)

def flush_lines_enter(state_globals):
    state_globals['external']['clear_sticky'] = True

def prepare_for_pretest_enter(state_globals):
    npxc.prepare_for_pretest_input(state_globals)



def scan_mouse_id_input(state_globals):
    """
    Input test function for state initialize
    """
    comp_id = os.environ.get('aibs_comp_id', socket.gethostname())
    mouse_id = state_globals["external"]["mouse_id"]
    user_id = state_globals["external"]["user_id"]
    
    if not isinstance(experiment, VariabilitySpontaneous):
        # now we can set the mouseID in mtrain 
        npxc.mtrain.mouse_id = mouse_id
        
    #@Ross do you want this only for experiments? I copied over for hab...
    logging.info(f'MID, {mouse_id}, UID, {user_id}, BID, {comp_id}, Received')

    state_globals["external"]["local_log"] = f'Mouse ID :{mouse_id}'
    state_globals["external"]["transition_result"] = True
    state_globals["external"]["status_message"] = "success"
    state_globals["external"]["clear_sticky"] = True

    try:
        result = limstk.donor_info(mouse_id)
        if "dummy_mode" in state_globals["external"] and (state_globals["external"]["dummy_mode"] == True):
            trigger_dir = npxc.config['dummy_trigger_dir']
        else:
            if "project" in result[0]["specimens"][0]:
                trigger_dir = result[0]["specimens"][0]["project"]["trigger_dir"]
            else:
                trigger_dir = state_globals["external"]["trigger_dir"]

        local_trigger_dir = npxc.config['local_trigger_dir']
        state_globals["external"]["dummy_mode"] = False

    except limstk.LIMSUnavailableError:
        message = f'Could not retrieve donor_info for {mouse_id} from LIMS.'
        npxc.fail_state(message, state_globals)
        return

    npxc.assess_previous_sessions(state_globals)
    session_dict = state_globals['external']['hab_sessions']
    session_count = npxc.count_sessions(session_dict, state_globals['external']['mouse_id'])
    guess_exp_day = 'normal'
    if session_count==0:
        guess_exp_day = 'first'
    elif (dt.today().weekday() == 1) and (session_count>3):
        guess_exp_day = '1_day_before_experiment'
    elif (dt.today().weekday() == 0) and (session_count>2):
        guess_exp_day ='2_days_before_experiment'
    state_globals['external']['entered_experiment_day'] = guess_exp_day

    state_globals["external"]["local_trigger_dir"] = local_trigger_dir
    state_globals["external"]["lims_specimen_id"] = result[0]["specimens"][0]["id"]
    state_globals["external"]["specimen.Name"] = result[0]["name"]
    state_globals["external"]["lims_project_id"] = result[0]["specimens"][0]["project_id"]
    state_globals["external"]["Project.Trigger_dir"] = trigger_dir

    if "project" in result[0]["specimens"][0]:
        state_globals["external"]["Project.Code"] = result[0]["specimens"][0]["project"].get("code", "No Project Code")
    else:
        state_globals["external"]["Project.Code"] = "No Project Code"

    state_globals["external"]["Project.Code1"] = state_globals["external"]["Project.Code"]
    state_globals["external"]["Project.Code2"] = state_globals["external"]["Project.Code"]
    state_globals["external"]["Project.Code3"] = state_globals["external"]["Project.Code"]
    state_globals["external"]["Project.Code4"] = state_globals["external"]["Project.Code"]

    # generate a session name
    session_name_timestamp = dt.now().strftime("%Y%m%d")
    session_name_string = f'{state_globals["external"]["mouse_id"]}_{session_name_timestamp}'
    logging.info(f'Session name: {session_name_string}')
    state_globals["external"]["session_name"] = session_name_string
    state_globals["external"]["sessionNameTimestamp"] = session_name_timestamp
    state_globals['external']['Auto.Date.String'] = dt.now().strftime("%Y%m%d")
    state_globals['external']['Auto.Date.String1'] = state_globals['external']['Auto.Date.String']
    state_globals['external']['next_state'] = 'LIMS_request'

    if 'use_auto' in state_globals['external'] and state_globals['external']['use_auto']:
        states_skipped = ['LIMS_request', 'date_string_check']
        fields_skipped = ['auto_generated_date_string', 'Project.Code.lims']
        npxc.skip_states(state_globals, states_skipped, fields_skipped, globals())
        state_globals['external']['next_state'] = 'check_experiment_day'




def LIMS_request_enter(state_globals):
    """
    Entry function for state initialize
    """
    state_globals['external']['clear_sticky'] = True
    state_globals['external']['Manual.Project.Code'] = state_globals['external']['Project.Code2']

def LIMS_request_input(state):
    """
    Input test function for state initialize
    """
    if 'Project.Code.new' in state['external'] and state['external']['Project.Code.new']:
        state['external']['Project.Code'] = state['external']['Manual.Project.Code']
    try:
        result = limstk.projects_validation(state["external"]["Project.Code"])
        if result:
            state["external"]["status_message"] = "success"
            state["external"][
                "local_log"
            ] = f'Project.Code:{state["external"]["Project.Code"]} validated'
        else:
            state["external"][
                "local_log"
            ] = f'Project.Code:{state["external"]["Project.Code"]} NOT validated'

            state["external"]["status_message"] = f"Failure:Project.Code Validation Error"

        state["external"]["transition_result"] = True
    except limstk.LIMSUnavailableError:
        npxc.fail_state(f'Error validating project code {state["external"]["Project.Code.LIMS"]} in LIMS', state)



def LIMS_request_revert(state_globals):
    ...

def check_stimulus_enter(state_globals):
    defaults = ['3']
    key_list = ['MTrainData', 'Habituation_Sessions', 'days_to_open_mtrain']
    days_to_open_mtrain = npxc.get_from_config(key_list, defaults)
    days_to_open_mtrain = [str(day_num) for day_num in days_to_open_mtrain]
    if (str(dt.today().weekday()+1) in days_to_open_mtrain) or ('first' in state_globals['external']['full_session_type']):
        message = "It looks like today is a day that requires an mtrain stage change. You should be extra careful to check the mtrain stage."
        npxc.alert_text(message, state_globals)
        # url = f'http://mtrain:5000/set_state/{state_globals["external"]["mouse_id"]}'
        # webbrowser.open(url)
    session_day = state_globals['external']['entered_experiment_day']
    mtrain_string = None
    try:
        default=None
        key_list = ['MTrainData', 'Habituation_Sessions', str(dt.today().weekday()+1), 'Full_String']
        mtrain_string = npxc.get_from_config(key_list, default)

        
    except Exception as E:
        pass
    if mtrain_string == None:
        key_list = ['MTrainData', 'Habituation_Sessions', 'Habituation', 'Full_String']
        default = 'VisualBehaviorEPHYS_Task1G_v0.0.10/HABITUATION_5_images_G_handoff_ready_3uL_reward '
        mtrain_string = npxc.get_from_config(key_list, default=default)

    state_globals['external']['mtrain_string'] = mtrain_string
    state_globals["external"]["clear_sticky"] = True


def date_string_check_enter(state_globals):
    state_globals['external']['Manual.Date.String'] = state_globals['external']['Auto.Date.String1']


def date_string_check_input(state):
    """
    Input test function for state initialize
    """
    print('>>>date_string_check_input<<<')
    if state["external"]["dummy_mode"]:
        result = limstk.isi_experiment_test(state["external"]["mouse_id"])
    else:
        result = limstk.isi_experiment_prod(state["external"]["mouse_id"])

    if not result:
        npxc.fail_state(f'Could not find an ISI experiment for mouse id:{state["external"]["mouse_id"]} failure', state)
        state["external"]["next_state"] = "lims_abort_confirm"
        return
    else:
        state["external"]["next_state"] = "check_experiment_day"

    isi_experiments_list = []
    target_list = []
    isi_id_list = []

    for item in result[0]["isi_experiments"]:
        isi_name = item["name"]
        state_local = item["workflow_state"]
        target_list = [item["targets"]]
        if len(target_list) >= 1:
            target_str = "Target Data Available"
        else:
            target_str = "No Target Data Found"

        isi_id = item["id"]
        experiment_str = f"{isi_name}   state:{state_local}   targets:{target_str}"
        isi_experiments_list.append(experiment_str)
        target_list.append(target_list)
        isi_id_list.append(isi_id)

    state["external"]["ISI_experiments"] = isi_experiments_list
    state["external"]["ISI_targets"] = target_list
    state["external"]["ISI_ids"] = isi_id_list
    target_path = f'/{result[0]["target_map_image_path"]}'
    overlay_path = f'/{result[0]["isi_image_overlay_path"]}'
    state["external"]["isi_image_path"] = target_path
    state["external"]["isi_overlay_path"] = overlay_path
    state["external"]["local_log"] = f'LIMS pull for isi experiments associated with mouse id:{state["external"]["mouse_id"]} success!'
    """
    Input test function for state initialize
    """
    # lots of LIMS interactions to be added here when service is available
    # find which experiment was selected:
    selected_index = 0
    #TODO - Have it actually select the ISI map that we put our coordinates on?
    #for x in range(len(state["external"]["ISI_experiments"])):
    #    if state["external"]["ISI_experiments"][x] == state["external"]["isi_experiment_selected"]:
    #        selected_index = x
    #        break

    project_name = state["external"].get("Project.Code", None)
    project_id = state["external"]["lims_project_id"]
    if project_name:
        try:
            url = "http://lims2/projects.json?code=" + project_name  # TODO: Configure this
            lims_result = requests.get(url)
            project_id = json.loads(lims_result.text)[0]["id"]
        except (KeyError, IndexError):
            npxc.fail_state(f"Error getting project id for project {project_name}", state)
            return

    state["external"]["selected_isi_id"] = state["external"]["ISI_ids"][selected_index]

    # create a unique name for the creating the ecephys session
    if 'auto_generated_date_string' in state['external'] and state['external']['auto_generated_date_string']:
        session_name_timestamp = state['external']['Auto.Date.String']
    else:
        session_name_timestamp = state['external']['Manual.Date.String']
    print(session_name_timestamp)
    assert(len(session_name_timestamp) == 8)
    session_name_string = f'{state["external"]["mouse_id"]}_{session_name_timestamp}'
    print('Session name string: '+session_name_string)
    state["external"]["session_name"] = session_name_string
    state["external"]["sessionNameTimestamp"] = session_name_timestamp


    timestamp = dt.now().strftime("%Y%m%d%H%M%S")
    name = f'HAB_{timestamp}_{state["external"]["lims_user_id"]}'
    request_json = {
        "specimen_id": state["external"]["lims_specimen_id"],
        "project_id": project_id,
        "isi_experiment_id": state["external"]["selected_isi_id"],
        "name": name,
        "operator_id": state["external"]["lims_user_id"],
    }

    url = "http://lims2/observatory/ecephys_session/create"
    response = requests.post(url, json=request_json)
    decoded_dict = json.loads(response.content.decode("utf-8"))
    ecephys_session_id = str(decoded_dict["id"])
    state["external"]["ecephys_session_id"] = ecephys_session_id
    print('Old session name: '+state["external"]["session_name"])
    print('ecephys_session_id: '+state["external"]["ecephys_session_id"])
    state["external"]["session_name"] = f'{ecephys_session_id}_{state["external"]["session_name"]}'
    print('New session name: '+state["external"]["session_name"])

    # create direction for session ID on local drive
    session_name_directory = os.path.join(
        state["external"]["local_lims_head"], state["external"]["session_name"]
    )
    state["external"]["session_name_directory"] = session_name_directory
    local_lims_location = session_name_directory
    print(f"local lims lication: {local_lims_location}")
    os.makedirs(local_lims_location, exist_ok=True)

    # try:
    #     notes_proxy = state["component_proxies"]["Notes"]
    #     notes_proxy.setID(str(state["external"]["mouse_id"]), str(state["external"]["session_name"]))
    #     notes_proxy.setNoSurgery(True)
    #     state["external"]["status_message"] = "success"
    #     state["external"]["component_status"]["Notes"] = True
    # except KeyError:
    #     npxc.fail_state('SurgeryNotes proxy is not defined.', state)
    #     state["external"]["component_status"]["Notes"] = False
    #     return
    # except Exception:
    #     npxc.fail_state('Error setting mouse and session name in SurgeryNotes', state)
    #     state["external"]["component_status"]["Notes"] = False
    #     return
    mapped_lims_location = f"{npxc.config['mapped_lims_location']}/{state['external']['session_name']}"
    state["external"]["mapped_lims_location"] = mapped_lims_location
    state["external"]["local_lims_location"] = local_lims_location
    state["external"]["transition_result"] = True
    state["external"]["status_message"] = "success"

def date_string_check_revert(state_globals):
    ...


def lims_abort_confirm_input(state):
    """
    Input test function for state initialize
    """
    state["external"]["transition_result"] = True
    state["external"]["status_message"] = "success"
    if "lims_abort_experiment" in state["external"] and state["external"]["lims_abort_experiment"]:
        state["external"]["next_state"] = "create_manifest_and_platform_json_and_sync_report"
    elif "lims_abort_cancel" in state["external"] and state["external"]["lims_abort_cancel"]:
        state["external"]["next_state"] = "pull_ISI_data"
    else:
        npxc.fail_state('No valid inputs', state)


def lims_abort_confirm_revert(state):
    ...


def check_experiment_day_enter(state_globals):
    defaults = ['3']
    key_list = ['MTrainData', 'Habituation_Sessions', 'days_to_open_mtrain']
    days_to_open_mtrain = npxc.get_from_config(key_list, defaults)
    days_to_open_mtrain = [str(day_num) for day_num in days_to_open_mtrain]
    # if (str(dt.today().weekday()+1) in days_to_open_mtrain) or ('first' in state_globals['external']['full_session_type']):
    #     url = f'http://mtrain:5000'
    #     webbrowser.open(url)

    state_globals["external"]["clear_sticky"] = True


def check_experiment_day_input(state_globals):
    entered_experiment_day = state_globals['external']['entered_experiment_day']
    session_types_options = state_globals['external']['session_types_options']
    if not (entered_experiment_day in session_types_options):
        message = 'Session Type Not Valid: please type exactly as listed'
        npxc.fail_state(message, state_globals)
    else:
        state_globals['external']['full_session_type'] = state_globals['external'][
                                                             'session_type'] + '_' + entered_experiment_day.lower()




def configure_hardware_enter(state_globals):
    """
    Entry function for state configure_hardware
    """
    print('>> configure_hardware_enter')

    # good place to test the components turning on/off

    # load the sync config file
    try:
        proxy = state_globals['component_proxies']['Sync']

        try:
            proxy.init()
            #proxy.load_config("C:/ProgramData/AIBS_MPE/sync/last.yml")
            state_globals['external']['status_message'] = 'success'
            state_globals['external']['component_status']["Sync"] = True
        except Exception as e:
            print(f'Sync load config failure:{e}!')
            state_globals['external']['status_message'] = f'Sync load config failure:{e}'
            state_globals['external']['component_status']["Sync"] = False
    except Exception as e:
        print(f'Sync proxy failure:{e}!')
        state_globals['external']['status_message'] = f'Sync proxy failure:{e}'
        state_globals['external']['component_status']["Sync"] = False

    # set the data file path for open ephys



def configure_hardware_input(state_globals):
    """
    Input test function for state configure_hardware
    """
    print('>>> in configure_hardware input')

    # find which experiment was selected:
    #for x in range(len(state_globals["external"]["ISI_experiments"])):
    #    if (state_globals["external"]["ISI_experiments"][x] == state_globals["external"]["isi_experiment_selected"]):
    #        selected_index = x
    #        break

    #selected_targets = state_globals['external']['ISI_targets'][selected_index]
    #print(f'selected_targets:{selected_targets}')
    # build up string of targets to display
    #probe_list = ['A', 'B', 'C', 'D', 'E', 'F']
    #for x in range(len(selected_targets)):
    #    print(f'target_data:{selected_targets[x]}')
    #    state_globals['external'][f'target_{probe_list[x]}_data'] = json.dumps(selected_targets[x])

    #state_globals['external']['next_state'] = 'load_mouse'
    state_globals['external']['transition_result'] = True
    state_globals['external']['status_message'] = 'success'

def load_mouse_enter(state_globals):
    npxc.load_mouse_enter(state_globals)

def load_mouse_input(state_globals):
    """
    Input test function for state load_mouse_headframe
    """
    # create a mouse in headframe timestamp (YYYYMMDDHHMMSS)
    #state_globals['external']['mouse_in_headframe_holder']
    state_globals['external']["HeadFrameEntryTime"] = dt.now().strftime('%Y%m%d%H%M%S')
    state_globals['external']['local_log'] = f'HeadFrameEntryTime:{state_globals["external"]["HeadFrameEntryTime"]}'
    #state_globals['external']['next_state'] = 'lower_probe_cartridge'
    message = npxc.load_mouse_behavior(state_globals)
    if message:
        npxc.fail_state(message, state_globals)
    state_globals['external']['transition_result'] = True
    state_globals['external']['status_message'] = 'success'
    npxc.save_platform_json(state_globals, manifest=False)

    state_globals["resources"]["final_depth_timer_start"] = dt.now()
    if 'day2' in state_globals['external']['full_session_type'].lower():
        npxc.alert_text('Remove extra agarose and ensure adequate silicon oil.', state_globals)


def lower_probe_cartridge_enter(state_globals):
    state_globals['external']['clear_sticky'] = True




def get_hab_wait_time(state_globals):
    wait_time = .1
    if '2_days_before_experiment' in state_globals['external']['full_session_type']:
        defaults = 600
        key_list = ['session_params', '2_days_before_experiment', 'wait_time']
        wait_time = npxc.get_from_config(key_list, defaults)
    elif '1_day_before_experiment' in state_globals['external']['full_session_type']:
        defaults = 900
        key_list = ['session_params', 'day1', 'message_1']
        wait_time = npxc.get_from_config(key_list, defaults)
    return wait_time


def get_hab_message_1(state_globals):
    message_1 = ''
    if '2_days_before_experiment' in state_globals['external']['full_session_type'].lower():
        defaults = ("Please have each probe descend 500um at 100um/min\n"
            "- Then retract to 0")
        key_list = ['session_params', '2_days_before_experiment', 'message_1']
        message_1 = npxc.get_from_config(key_list, defaults)
    elif '1_day_before_experiment' in state_globals['external']['full_session_type'].lower():
        defaults = ("Please tap the coverslip with the probes close to the targets\n"
            "- Take a screenshot of the coordintes \n\n"

            f"- Adjust the zoom to {state_globals['external']['high_zoom_level']}\n\n"
            "- Then retract them to 0 and have each porbe descend 500um at 100um/min\n"
            )
        key_list = ['session_params', '1_day_before_experiment', 'message_1']
        message_1 = npxc.get_from_config(key_list, defaults)
    if message_1:
        npxc.alert_text(message_1, state_globals)
    return message_1

def get_hab_message_2(state_globals):
    message_2 = ''
    if '1_day_before_experiment' in state_globals['external']['full_session_type'].lower():
        defaults = ("- Please run probeLocator on the image from todays tap\n\n"
            "- Please check the perfusions requests for this week\n")
        key_list = ['session_params', '1_day_before_experiment', 'message_2']
        message_2 = npxc.get_from_config(key_list, defaults)
    if message_2:
        npxc.alert_text(message_2, state_globals)
    return message_2

def get_hab_message_3(state_globals):
    message_1 = ''
    session_type = state_globals['external']['full_session_type']
    if ('2_days_before_experiment' in session_type) or ('1_day_before_experiment' in session_type):
        defaults = ("Please retract the probes to Z height 0 um\n")
        key_list = ['session_params', 'day_before_experiment', 'message_3']
        message_1 = npxc.get_from_config(key_list, defaults)
    if message_1:
        npxc.alert_text(message_1, state_globals)
    return message_1


def get_hab_message_4(state_globals):
    message_1 = ''
    if (dt.today().weekday() == 2) or (dt.today().weekday() == 3):
        defaults = ("Please clean the eytracking mirror in preperation for the afternoons\n")
        key_list = ['session_params', 'day_before_experiment', 'message_3']
        message_1 = npxc.get_from_config(key_list, defaults)
    if message_1:
        npxc.alert_text(message_1, state_globals)
    return message_1


def pre_stimulus_wait_enter(state_globals):
    wait_time = get_hab_wait_time(state_globals)
    npxc.settle_timer_enter(state_globals, wait_time)
    #get_hab_message_1(state_globals)
    state_globals["external"]["clear_sticky"] = True


def pre_stimulus_wait_input(state_globals):
    """
    Input test function for state pre_stimulus_wait
    """
    # before transiting to the next state, get the listing for camstim files on the stim computer
    # recreate the proxy
    stim_files = []

    state_globals['external']['next_state'] = 'prime_lickspout'

    wait_time = get_hab_wait_time(state_globals)
    npxc.settle_timer_enter(state_globals, wait_time)

    if state_globals["external"].get("override_settle_timer", False):
        return  # TODO: make the checkbox optional

    if state_globals['external']['settle_time_remaining_num']:  # total_seconds < npxc.config['final_depth_timer_s']:
        message = 'The settle time has not elapsed! Please wait until the state timer matches the remaining time'
        npxc.fail_state(message, state_globals)
        return

    # state_globals['external']['next_state'] = 'check_data_dirs'
    state_globals["external"]["transition_result"] = True
    state_globals["external"]["status_message"] = "success"
    npxc.save_platform_json(state_globals, manifest=False)


def lower_probe_cartridge_input(state_globals):
    """
    Input test function for state lower_probe_cartridge
    """
    # create a probe_cartridge_lower timestamp (YYYYMMDDHHMMSS)
    if state_globals['external']['probe_cartridge_lower']:
        state_globals['external']["CartridgeLowerTime"] = dt.now().strftime('%Y%m%d%H%M%S')
        state_globals['external']['local_log'] = f'CartridgeLowerTime: {state_globals["external"]["CartridgeLowerTime"]}'
    try:
        npxc.rename_mlog(state_globals)
    except Exception as e:
        print(f"Notes rename log failure:{e}!")
    #state_globals['external']['next_state'] = 'brain_surface_focus'
    state_globals['external']['transition_result'] = True
    state_globals['external']['status_message'] = 'success'


def brain_surface_focus_enter(state_globals):
    state_globals['external']['clear_sticky'] = True
    state_globals['external']['old_low_zoom_level'] = state_globals['external']['low_zoom_level']
    state_globals['external']['low_zoom_level'] = state_globals['external']['high_zoom_level']
    get_hab_message_1(state_globals)


def brain_surface_focus_input(state_globals):
    """
    Input test function for state brain_surface_focus
    """
    # take a photo here from CamViewer3d
    if state_globals['external']['dummy_mode']:
        surface_1_left_path = os.path.join(state_globals['external']['mapped_lims_location'], (state_globals['external']["session_name"] + '_surface-image2-left.png'))
        surface_1_right_path = os.path.join(state_globals['external']['mapped_lims_location'], (state_globals['external']["session_name"] + '_surface-image2-right.png'))
        surface_1_local_path = os.path.join(state_globals['external']['local_lims_location'], (state_globals['external']["session_name"] + '_surface-image2-left.png'))
        surface_1_left_local_path = os.path.join(state_globals['external']['local_lims_location'], (state_globals['external']["session_name"] + '_surface-image2-left.png'))
        surface_1_right_local_path = os.path.join(state_globals['external']['local_lims_location'], (state_globals['external']["session_name"] + '_surface-image2-right.png'))
    else:
        surface_1_left_path = f'{state_globals["external"]["mapped_lims_location"]}/{state_globals["external"]["session_name"]}_surface-image2-left.png'
        surface_1_right_path = f'{state_globals["external"]["mapped_lims_location"]}/{state_globals["external"]["session_name"]}_surface-image2-right.png'
        surface_1_local_path = f'{state_globals["external"]["local_lims_location"]}/{state_globals["external"]["session_name"]}_surface-image2-left.png'
        surface_1_left_local_path = os.path.join(state_globals['external']['local_lims_location'], (state_globals['external']["session_name"] + '_surface-image2-left.png'))
        surface_1_right_local_path = os.path.join(state_globals['external']['local_lims_location'], (state_globals['external']["session_name"] + '_surface-image2-right.png'))
    try:

        print('>>>>>>> brain_surface_image')
        print(f'surface_1_left_path:{surface_1_left_path}')
        print(f'surface_1_right_path:{surface_1_right_path}')
        print(f'surface_1_local_path:{surface_1_local_path}')
        print('<<<<<<<')

        state_globals['external']['surface_1_left_name'] = f'{state_globals["external"]["session_name"]}_surface-image1-left.png'
        state_globals['external']['surface_1_right_name'] = f'{state_globals["external"]["session_name"]}_surface-image1-right.png'

        try:
            npxc.take_left_snapshot(state_globals,surface_1_left_path)
            npxc.take_right_snapshot(state_globals,surface_1_right_path)

            state_globals['external']['status_message'] = 'success'
            state_globals['external']['local_log'] = f'Surface_1_Path:{surface_1_local_path}'
        except Exception as e:
            print(f'Cam3d take photo failure:{e}!')
            state_globals['external']['status_message'] = f'Cam3d take photo failure:{e}'
            state_globals['external']['component_status']["Cam3d"] = False
    except Exception as e:
        print(f'Cam3d proxy failure:{e}!')
        state_globals['external']['status_message'] = f'Cam3d proxy failure:{e}'
        state_globals['external']['component_status']["Cam3d"] = False

    # check for the image files...make sure they were taken succesfully
    left_image_result = os.path.isfile(surface_1_left_local_path)
    right_image_result = os.path.isfile(surface_1_right_local_path)

    image_error_message ="Image Error:"

    if not left_image_result:
        image_error_message = image_error_message + " Left Image Not Taken!"

    if not right_image_result:
        image_error_message = image_error_message + " Right Image Not Taken!"

    if (not left_image_result) or (not right_image_result): # if one of the images not take successfully, force the warning box
        state_globals["external"]["alert"] = {'msg_text': image_error_message, 'severity':'Critical', 'informative_text': 'Check Cam3d Viewer, and restart if necessary.  Retake Image'}

    state_globals['external']['local_log'] = f'surface_1_path:{surface_1_local_path}'
    state_globals['external']['surface_1_file_location'] = surface_1_local_path # this is what gets displayed in the GUI
    if not(os.path.exists(surface_1_local_path)):
        time.sleep(5)
        if not(os.path.exists(surface_1_local_path)):
            message = 'You may need to click the blue button and blue triangle on Cam3d or restart it, Please also confirm there is only one camviewer gui open'
            npxc.alert_text(message, state_globals)
    state_globals['external']['surface_1_left_local_file_location'] = surface_1_left_local_path
    state_globals['external']['surface_1_right_local_file_location'] = surface_1_right_local_path

    #state_globals['external']['next_state'] = 'photodoc_confirm'
    state_globals['external']['transition_result'] = True
    state_globals['external']['status_message'] = 'success'


def photodoc_confirm_input(state_globals):
    """
    Input test function for state lower_probes_automatically
    """
    # create a insert_probes timestamp (YYYYMMDDHHMMSS)
    if 'retake_image' in state_globals['external'] and state_globals['external']['retake_image']:
        state_globals['external']['next_state'] = 'brain_surface_focus'
        state_globals['external'].pop('retake_image', None)
    else:
        state_globals['external']['next_state'] = 'confirm_ISI_match'
    state_globals['external']['transition_result'] = True
    state_globals['external']['status_message'] = 'success'


def confirm_ISI_match_enter(state_globals):
    if '1_day_before_experiment' in state_globals['external']['full_session_type'].lower():
        message = f"- Adjust the zoom to {state_globals['external']['old_low_zoom_level']}"
        npxc.alert_text(message, state_globals)

def select_stimulus_enter(state_globals):
    """
    Entry function for state select_stimulus
    """
    stim_files=[]


    state_globals['external']['local_log'] = f'Available Stim Files:{stim_files}'
    state_globals['external']['stim_files'] = stim_files



def select_stimulus_input(state_globals):
    """
    Input test function for state select_stimulus
    """
    print(f'stim file selected:{state_globals["external"]["stimulus_selected"]}')
    state_globals['external']['local_log'] = f'stim file selected:{state_globals["external"]["stimulus_selected"]}'

    #state_globals['external']['next_state'] = 'initiate_experiment'
    state_globals['external']['status_message'] = 'success'
    state_globals['external']['transition_result'] = True


def initiate_experiment_input(state_globals):
    """
    Input test function for state initiate_experiment
    """
    print('>> initiate_experiment_input <<')

    # recreate the proxy
    # create a experiment start time timestamp (YYYYMMDDHHMMSS)

    state_globals['external']["ExperimentStartTime"] = dt.now().strftime('%Y%m%d%H%M%S')
    state_globals['external']['local_log'] = f'ExperimentStartTime:{state_globals["external"]["ExperimentStartTime"]}'
    state_globals['external']['sync_file_path'] = f'{state_globals["external"]["local_lims_location"]}/{state_globals["external"]["session_name"]}.sync'

    state_globals['external']['sync_temp_file'] = f'C:/ProgramData/AIBS_MPE/sync/output/{state_globals["external"]["session_name"]}_temp'

    try:
        proxy = state_globals['component_proxies']['Sync']
        try:
            proxy.start(path=state_globals['external']['sync_temp_file'])
            state_globals['external']['sync_temp_file'] = f'C:/ProgramData/AIBS_MPE/sync/output/{state_globals["external"]["session_name"]}_temp.h5'
            state_globals['external']['status_message'] = 'success'
        except Exception as e:
            print(f'Sync start failure:{e}!')
            state_globals['external']['status_message'] = f'Sync start failure:{e}'
            state_globals['external']['component_status']["Sync"] = False
    except Exception as e:
        print(f'Sync proxy failure:{e}!')
        state_globals['external']['status_message'] = f'Sync proxy failure:{e}'
        state_globals['external']['component_status']["Sync"] = False


    # put in a delay between the sync starting and the videomon starting
    time.sleep(3)

    try:
        proxy = state_globals['component_proxies']['VideoMon']
        try:
            state_globals['external']['behavior_file_path'] = f'{state_globals["external"]["mapped_lims_location"]}/{state_globals["external"]["session_name"]}.behavior.avi'
            state_globals['external']['behavior_local_file_path'] = f'{state_globals["external"]["local_lims_location"]}/{state_globals["external"]["session_name"]}.behavior.avi'
            state_globals['external']['behavior_local_file_name'] = f'{state_globals["external"]["session_name"]}.behavior.avi'

            state_globals['external']['eyetracking_file_path'] = f'{state_globals["external"]["mapped_lims_location"]}/{state_globals["external"]["session_name"]}.eye.avi'
            state_globals['external']['eyetracking_local_file_path'] = f'{state_globals["external"]["local_lims_location"]}/{state_globals["external"]["session_name"]}.eye.avi'
            state_globals['external']['eyetracking_local_file_name'] = f'{state_globals["external"]["session_name"]}.eye.avi'

            state_globals['external']['videomon_file_path'] = f'C:/ProgramData/AIBS_MPE/videomon/data/{state_globals["external"]["session_name"]}'
            print(f'videomon file path:{state_globals["external"]["videomon_file_path"]}')
            proxy.start_record(path=state_globals["external"]["videomon_file_path"])
            state_globals['external']['status_message'] = 'success'
        except Exception as e:
            logging.exception(f'VideoMon start failure:')
            print(f'VideoMon start failure:{e}!')
            state_globals['external']['status_message'] = f'Videomon start failure:{e}'
            state_globals['external']['component_status']["VideoMon"] = False
    except Exception as e:
        print(f'VideoMon proxy failure:{e}!')
        state_globals['external']['status_message'] = f'Videomon proxy failure:{e}'
        state_globals['external']['component_status']["VideoMon"] = False

    # start the stim process
    # recreate the proxy
    try:
        proxy = state_globals['component_proxies']['Stim']
        try:
            print(f'starting stim:{state_globals["external"]["stimulus_selected"]}')
            proxy.start_script_from_path(state_globals["external"]["stimulus_selected"])
            state_globals['external']['status_message'] = 'success'
        except Exception as e:
            print(f'Stim start failure:{e}!')
            state_globals['external']['status_message'] = f'Stim start failure:{e}'
            state_globals['external']['component_status']["Stim"] = False
    except Exception as e:
        print(f'Stim proxy failure:{e}!')
        state_globals['external']['status_message'] = f'Stim proxy failure:{e}'
        state_globals['external']['component_status']["Stim"] = False

    state_globals['external']['clear_sticky'] = True
    #state_globals['external']['next_state'] = 'experiment_running_timer'
    state_globals['external']['transition_result'] = True
    state_globals['external']['status_message'] = 'success'
    npxc.save_platform_json(state_globals, manifest=False)


def experiment_running_timer_enter(state_globals):
    get_hab_message_2(state_globals)
    if 'first' in state_globals['external']['full_session_type']:
        message = ("It looks like its the first session with this mouse. "
        "You should be extra careful to check the lickspout position. "
        "Please watch for chin hits, paw hits, "
        "compulsive licking and first lick misses.\n"
        "Adjust spout position and wheel height accordingly.")
        npxc.alert_text(message, state_globals)


def monitor_experiment_input(state_globals):
    npxc.monitor_experiment_input(state_globals)


def initiate_behavior_stimulus_input(state_globals):
    npxc.initiate_behavior_stimulus_input(state_globals)
    npxc.save_platform_json(state_globals, manifest=False)

    print('>> initiate_behavior_experiment_input <<')


def camstim_ping_error_input(state_globals):
    npxc.camstim_ping_error_input(state_globals)

    print('>> camstim_ping_error_input <<')


def end_experiment_enter(state_globals):
    get_hab_message_3(state_globals)

def end_experiment_input(state_globals):
    """
    Input test function for state end_experiment
    """

    # get an experiment end time
    state_globals["external"]["ExperimentCompleteTime"] = dt.now().strftime("%Y%m%d%H%M%S")
    state_globals["external"][
        "local_log"
    ] = f'ExperimentCompleteTime:{state_globals["external"]["ExperimentCompleteTime"]}'


    # end the stim process
    #opto_stim_file_path = os.path.join(
    #    state_globals["external"]["local_lims_location"], state_globals["external"]["session_name"] + ".opto.pkl"
    #)
    #visual_stim_file_path = os.path.join(
    #    state_globals["external"]["local_lims_location"], state_globals["external"]["session_name"] + ".stim.pkl"
    #)

    #state_globals["external"]["opto_stim_file_name"] = f'{state_globals["external"]["session_name"]}.opto.pkl'
    #state_globals["external"]["visual_stim_file_name"] = f'{state_globals["external"]["session_name"]}.stim.pkl'

    #state_globals["external"]["opto_stim_file_location"] = opto_stim_file_path
    #state_globals["external"]["visual_stim_file_location"] = visual_stim_file_path

    #print(
    #    f'mapped_lims_location:{state_globals["external"]["mapped_lims_location"]}, opto file name:{state_globals["external"]["opto_stim_file_name"]}'
    #)

    # recreate the proxy



    # stop the open ephys process
    #npxc.stop_ecephys_recording(state_globals)

    # stop the VideoMon process


    time.sleep(3)
    npxc.stop_common_session_monitoring(state_globals)
    # recreate the proxy

    # state_globals['external']['next_state'] = 'remove_probes_start'
    state_globals["external"]["transition_result"] = True
    state_globals["external"]["status_message"] = "success"

    state_globals["external"][
        "local_log"
    ] = f'ExperimentCompleteTime:{state_globals["external"]["ExperimentCompleteTime"]}'
    state_globals["external"]["transition_result"] = True
    npxc.save_platform_json(state_globals, manifest=False)


def initiate_behavior_habituation_input(state_globals):
    """
    Input test function for state initiate_experiment
    """

    # recreate the proxy
    # create a experiment start time timestamp (YYYYMMDDHHMMSS)

    print('>> initiate_behavior_habituation_input <<')

    npxc.start_common_session_monitoring(state_globals)

    state_globals['external']['next_statet'] = 'experiment_running_timer'
    initiate_behavior_stimulus_input(state_globals)

    #state_globals['external']['clear_sticky'] = True
    #state_globals['external']['next_state'] = 'experiment_running_timer'
    state_globals['external']['transition_result'] = True
    state_globals['external']['status_message'] = 'success'


def initiate_passive_habituaiton_input(state_globals):
    """
    Input test function for state initiate_experiment
    """
    # recreate the proxy
    # create a experiment start time timestamp (YYYYMMDDHHMMSS)


    print('>> initiate_passive_habituaiton_input <<')


    npxc.start_common_session_monitoring(state_globals)
    npxc.start_stim(state_globals)
    #state_globals['external']['clear_sticky'] = True
    #state_globals['external']['next_state'] = 'experiment_running_timer'
    state_globals['external']['transition_result'] = True
    state_globals['external']['status_message'] = 'success'


def remove_mouse_and_move_files2_input(state_globals):
    print('>> remove_mouse_and_move_files2_input <<')
    message = npxc.remove_mouse_input(state_globals)
    if message:
        npxc.fail_state(message, state_globals)

    ###Can't do this without retrieving water earned which we don't do...
    predict_earned = npxc.get_from_config(['predict_earned'], default=True)
    if predict_earned:
        try:
            pre_weight = float(state_globals['external']['mouse_weight_pre'])
            post_weight = float(state_globals['external']['mouse_weight_post'])
            loss_per_hour = float(npxc.get_from_config(['mouse_weight_loss_per_hr'], default=0))
            loss_per_session = float(npxc.get_from_config(['mouse_weight_loss_per_session'], default=.3))
            try:
                enter_time = int(state_globals["external"]["HeadFrameEntryTime"])
                exit_time = int(state_globals["external"]["HeadFrameExitTime"])
                experient_length_hrs = (exit_time - enter_time)/10000 #(approximate hours since I don't want to bother converting back to a real dt)

            except Exception as E:
                experient_length_hrs = float(npxc.get_from_config(['expected_hab_time'], default=1.2))

            print(f'approximate experiment length in hours: {experient_length_hrs}')
            expected_loss = experient_length_hrs*loss_per_hour
            expected_earned = round(((post_weight+ expected_loss +loss_per_session) - pre_weight), 2)
            state_globals['external']['expected_earned'] = expected_earned
            message = (f"Based on the pre and post weight we expect that the mouse earned {expected_earned} mls of water. "
            "If this is significantly wrong the solonoid may need to be calibrated")
            #npxc.alert_text(message, state_globals)
        except Exception as E:
            message = 'Failure occured trying to predict water earned'
            npxc.alert_text(message, state_globals)

    npxc.save_platform_json(state_globals, manifest=False)

def water_mouse_enter(state_globals):
    npxc.water_mouse_enter(state_globals)


def water_mouse_input(state_globals):
    npxc.water_mouse_input(state_globals)


def water_mouse_exit(state_globals):
    npxc.water_mouse_exit(state_globals)
    npxc.save_platform_json(state_globals, manifest=False)


def check_files1_enter(state_globals):
    get_hab_message_4(state_globals)

def check_files1_input(state_globals):
    print('>> check_files1_input <<')

    session_type = state_globals['external']['full_session_type']
    pkl_keyword = 'behavior-'
    behavior_pkl_path = npxc.get_pkl_path(pkl_keyword, state_globals)
    npxc.overwrite_foraging_id(behavior_pkl_path, session_type, state_globals)
    checkpoint = 1
    npxc.videomon_copy_wrapup(state_globals)
    if '1_day' in state_globals['external']['full_session_type']:
        session_type = state_globals['external']['full_session_type']
    else:
        session_type = 'behavior_habituation'
    failed = npxc.check_files_input(state_globals, session_type, checkpoint)

    if failed:
        #state_globals['external']['next_state'] = 'files_error1'
        npxc.alert_from_error_dict(state_globals, failed, primary_key=False)
    validation_type = state_globals['external']['session_type']
    validation_type = state_globals['external']['full_session_type'] + '_lims_ready'
    npxc.run_validation(state_globals, validation_type)
    failed = npxc.get_validation_results(state_globals, validation_type)
    state_globals['external']['next_state'] = 'create_manifest_and_platform_json_and_sync_report'
    if failed:
        foraging_failed = False
        for key, message in failed.items():
            if 'foraging' in key or 'foraging' in message:
                foraging_failed = True
        if foraging_failed:
            npxc.overrideable_error_state(state_globals, 'check_files1', message=message)
            state_globals['external']['next_state'] = 'foraging_ID_error'
        else:
            state_globals['external']['next_state'] = 'files_error1'
            npxc.alert_from_error_dict(state_globals, failed)
    npxc.save_platform_json(state_globals, manifest=False)

    print(f'Missing files is {failed}')


@npxc.state_transition
def foraging_ID_error_input(state_globals):
    #print('foraging ID:' + state_globals['external']['foraging_id'])
    #print('stimulus_name:' + state_globals['external']['stimulus_name'])
    #print('script_name:' + state_globals['external']['script_name'])
    #npxc.overrideable_error_state_input(state_globals, 'water_mouse', 'check_files1')
    pass


@npxc.state_transition
def foraging_ID_error_exit(state_globals):
    pass
    #npxc.overrideable_error_state_exit(state_globals)


def files_error1_input(state_globals):
    print(">> files_error_input <<")
    if "check_files_retry" in state_globals["external"] and state_globals["external"]["check_files_retry"]:
        state_globals["external"].pop("check_files_retry", None)
        check_files1_input(state_globals)
    else:
        state_globals['external']['next_state'] = 'create_manifest_and_platform_json_and_sync_report'
    state_globals['external']['transition_result'] = True
    state_globals['external']['status_message'] = 'success'

def create_manifest_and_platform_json_and_sync_report_input(state_globals):
    """
    Input test function for state create_manifest_and_platform_json_and_sync_report
    """
    # LIMS Stuff will wait until I get the LIMS Session stuff sorted out with RH
    npxc.save_platform_json(state_globals, manifest=True)
    print("going to workflow complete...")
    state_globals["external"]["transition_result"] = True
    state_globals["external"]["status_message"] = "success"


@npxc.state_transition
def cleanup_enter(state_globals):
    """
    Entry function for state workflow_complete
    """
    print(f">> Cleanup enter!")

    # if aborting, make sure to stop capture and release cameras in Cam3d



@npxc.state_transition
def cleanup_input(state_globals):
    """
    Input test function for state workflow_complete
    """
    print(f">> Cleanup Input <<")
    if'stop_streams' in state_globals['external'] and state_globals['external']['stop_streams']:
        npxc.stop_common_session_monitoring(state_globals)

    npxc.save_platform_json(state_globals, manifest=False)
    state_globals["external"]["next_state"] = "workflow_complete"
    state_globals["external"]["transition_result"] = True
    state_globals["external"]["status_message"] = "success"


def check_files2_input(state_globals):
    print('>> check_files2_input <<')
    session_type = state_globals['external']['session_type']
    checkpoint = 2
    stop = npxc.check_files2(state_globals, session_type, checkpoint)
    print(f'stop is {stop}')
    if stop:
        state_globals['external']['next_state'] = 'files_error2'
    else:
        state_globals['external']['next_state'] = 'copy_files_to_network'
    state_globals['external']['transition_result'] = True
    state_globals['external']['status_message'] = 'success'


def files_error2_input(state_globals):
    print('>> files_error2_input <<')
    if 'check_files2_retry' in state_globals['external'] and state_globals['external']['check_files2_retry']:
        print('go back to initialize')
        state_globals['external']['next_state'] = 'check_files2'
        state_globals['external'].pop('check_files2_retry', None)
    elif 'create_files2_retry' in state_globals['external'] and state_globals['external']['create_files2_retry']:
        print('go back to initialize')
        state_globals['external']['next_state'] = 'create_manifest_and_platform_json_and_sync_report'
        state_globals['external'].pop('create_files2_retry', None)
    else:
        state_globals['external']['next_state'] = 'initiate_data_processing'
    state_globals['external']['transition_result'] = True
    state_globals['external']['status_message'] = 'success'

def copy_files_to_network_input(state_globals):
    """
    Entry function for state workflow_complete
    """
    print(f'>> copy_files_to_network_enter <<')
    npxc.save_notes(state_globals)
    time.sleep(1)
    npxc.backup_files(state_globals)
    npxc.probes_need_cleaning(state_globals)

    if dt.today().weekday() == 2:
        message = ("It looks like there may be an experiment later today. "
            "Please make sure to dip the probes in MilliQ water")
        npxc.alert_text(message, state_globals)
        probes_need_cleaning = True
    print('enter workflow complete')
    logging.start_stop('Neuropixels workflow complete', extra={'weblog': True})

def ready_to_check_network_input(state_globals):
    print('>> copy_files_to_network input <<')
    npxc.save_notes(state_globals)
    session_type = state_globals['external']['session_type']
    if npxc.global_processes['network_backup_process'].poll() is None:
        message = 'Files are not finished copying to the network. Wait a bit and try again'
        npxc.fail_state(message, state_globals)
    else:
        failed = {}
        failed = npxc.check_files_network(state_globals, session_type, {1, 2})
        print(f"missing_files network is {failed}")
        validation_type = state_globals['external']['full_session_type'] + '_local_qc'
        npxc.run_validation(state_globals, validation_type)
        failed.update(npxc.get_validation_results(state_globals, validation_type))
        state_globals['external']['next_state'] = 'create_manifest_and_platform_json'
        if failed:
            npxc.alert_from_error_dict(state_globals, failed)
            state_globals["external"]["next_state"] = "network_backup_error"
        else:
            state_globals["external"]["next_state"] = "workflow_complete"
        state_globals["external"]["transition_result"] = True
        state_globals["external"]["status_message"] = "success"

def network_backup_error_input(state_globals):
    print('>> files_error2_input <<')
    if 'copy_retry' in state_globals['external'] and state_globals['external']['copy_retry']:
        print('go back to initialize')
        state_globals['external']['next_state'] = 'copy_files_to_network'
        state_globals['external'].pop('copy_retry', None)
    elif 'data_retry' in state_globals['external'] and state_globals['external']['data_retry']:
        print('go back to initialize')
        state_globals['external']['next_state'] = 'ready_to_check_network'
        state_globals['external'].pop('data_retry', None)
    else:
        state_globals['external']['next_state'] = 'workflow_complete'
    state_globals['external']['transition_result'] = True
    state_globals['external']['status_message'] = 'success'

def workflow_complete_enter(state_globals):
    """
    Entry function for state workflow_complete
    """
    logging.start_stop('Neuropixels workflow complete', extra={'weblog': True})
    state_globals['workflow_complete'] = True

def workflow_complete_input(state_globals):
    """
    Input test function for state workflow_complete
    """
    print("workflow complete input")
    state_globals["external"]["workflow_complete_time"] = dt.now().strftime('%Y%m%d%H%M%S')
    npxc.save_platform_json(state_globals, manifest=False)

    state_globals["external"]["next_state"] = None
    state_globals["external"]["transition_result"] = True
    state_globals["external"]["status_message"] = "success"



from __future__ import annotations

import contextlib
import datetime
import logging
import pathlib
import time
import logging
import math
import os
import shutil
import subprocess
import sys
from typing import Any, Generator, Literal, Mapping, Optional, Sequence, Type

import np_config
import np_logging 

from np_workflows.services import zro
from .config import Rig, RIG_ID
from .protocols import Stoppable

logger = np_logging.getLogger(__name__)

def config_from_zk(rig: Optional[Literal[0, 1, 2]]=None) -> Mapping[str, Any]:
    " Common `services` config plus rig-specific `services`"
    common_config = np_config.from_zk('/projects/np_workflows/defaults/configuration')['services']
    
    if rig:
        rig_config = np_config.from_zk(f'/rigs/NP.{rig}')['services']
    else:
        rig_config = np_config.from_zk(f'/rigs/{RIG_ID}')['services']

    for k, v in rig_config.items():
        common_config[k] = common_config.get(k, {}) | v
    return common_config

def start_rsc_app(host, app_id) -> None:
    rsc_node = zro.Proxy(host, 6000)
    if rsc_node.p_status().get(app_id) == 0:
        logger.info("Launching %s on %s via RSC", app_id, host)
        rsc_node.p_start(app_id)
        time.sleep(.1)
        if rsc_node.p_status().get(app_id) != 0:
            logger.warning("Failed to start %s", app_id)
            return
    logger.info("%s is running on %s", app_id, host)
            
@contextlib.contextmanager
def debug_logging() -> Generator[None, None, None]:
    root_logger = logging.getLogger()
    level_0 = root_logger.level
    root_logger.setLevel(logging.DEBUG)
    try:
        yield
    finally:
        root_logger.setLevel(level_0)

@contextlib.contextmanager
def stop_on_error(obj: Stoppable, reraise=True):
    if not isinstance(obj, Stoppable):
        raise TypeError(f"{obj} does not support stop()")
    try:
        yield
    except Exception as exc:
        with contextlib.suppress(Exception):
            obj.exc = exc
            logger.info("%s interrupted by error:", obj.__name__, exc_info=exc)
            obj.stop()
        if reraise:
            raise exc

@contextlib.contextmanager
def suppress(*exceptions: Type[BaseException]):
    try:
        yield
    except exceptions or Exception as exc:
        with contextlib.suppress(Exception):        
            logger.error("Error suppressed: continuing despite raised exception", exc_info=exc)
    finally:
        return
    
def is_online(host: str) -> bool:
    "Use OS's `ping` cmd to check if `host` is online."
    command = ["ping", "-n" if "win" in sys.platform else "-c", "1", host]
    try:
        return subprocess.call(command, stdout=subprocess.PIPE, timeout=1.0) == 0
    except subprocess.TimeoutExpired:
        return False
    
def unc_to_local(path: pathlib.Path) -> pathlib.Path:
    "Convert UNC path to local path if on Windows."
    if 'win' not in sys.platform:
        return path
    comp = os.environ["COMPUTERNAME"]
    if comp in path.drive:
        drive = path.drive.split('\\')[-1]
        drive = drive[:-1] if drive[-1] == '$' else drive
        drive = drive + ':' if drive[-1] != ':' else drive 
        drive += '\\'
        path =  pathlib.Path(drive, path.relative_to(path.drive))
    return path

def free_gb(path: str|bytes|os.PathLike) -> float:
    "Return free space at `path`, to .1 GB. Raises FileNotFoundError if `path` not accessible."
    path = pathlib.Path(path)
    path = unc_to_local(path)
    return round(shutil.disk_usage(path).free / 1e9, 1)

def get_files_created_between(
    path: str | bytes | os.PathLike, 
    glob: str = '*',
    start: float | datetime.datetime = 0, 
    end: float | datetime.datetime = None,
) -> list[pathlib.Path]:
    path = pathlib.Path(path)
    if not path.is_dir():
        path = path.parent
    if not end:
        end = time.time()
    start = start.timestamp() if isinstance(start, datetime.datetime) else start
    end = end.timestamp() if isinstance(end, datetime.datetime) else end
    ctime = lambda x: x.stat().st_ctime
    files = (file for file in path.glob(glob) if int(start) <= ctime(file) <= end)
    return sorted(files, key=ctime)

def is_file_growing(path: str|bytes|os.PathLike) -> bool:
    "Compares size of most recent .sync data file at two time-points - will block for up to 20s depending on file-size."
    path = pathlib.Path(path)
    size_0 = path.stat().st_size
    # for sync: file is appended periodically in chunks that scale non-linearly with size
    if '.sync' == path.suffix:
        time.sleep(2 * math.log10(size_0))
    else:
        time.sleep(.5 * math.log10(size_0))
    if path.stat().st_size == size_0:
        return False
    return True



from __future__ import annotations

import json
import logging
import os
import pathlib
import shutil
import socket
import sys
import time
import enum
from typing import Any, Optional, Sequence

import requests
import np_logging

from np_workflows.services import utils, config, protocols, zro
from np_workflows.services.protocols import TestFailure

# global vars -------------------------------------------------------------------------- #
logger = np_logging.getLogger(__name__) # logs will show full module path

__name__ = "OpenEphys" # Service protocol operations will see just the 'class' name

exc: Optional[BaseException] = None
initialized: float = 0
"`time.time()` when the service was initialized."

host: str = config.Rig.Acq.host
port: str | int = 37497 # 1-800-EPHYS
latest_start: float = 0
"`time.time()` when the service was last started via `start()`."

# for launching:
rsc_app_id = 'open-ephys'

# device records:
gb_per_hr: int | float = 250 # per drive
min_rec_hr: int | float = 2
pretest_duration_sec: int | float = .5

# for resulting data:
folder: str #! required
"The string that will be sent to Open Ephys to name the recording: typically `0123456789_366122_20220618`"
data_files: Sequence[pathlib.Path] = []
"Storage for paths collected over the experiment."
data_root: Optional[pathlib.Path] = None
# -------------------------------------------------------------------------------------- #

class State(enum.Enum):
    idle = "IDLE"
    acquire = "ACQUIRE"
    record = "RECORD"

class Endpoint(enum.Enum):
    status = "status"
    recording = "recording"
    processors = "processors"
    message = "message"
    
def launch() -> None:
    utils.start_rsc_app(host, rsc_app_id)
        
def pretest() -> None:
    logger.info("OpenEphys | Starting pretest")
    global folder
    folder = '_pretest_'
    initialize()
    test()
    try:
        start()
        time.sleep(pretest_duration_sec)
        verify()
    finally:
        stop()
    finalize()
    validate()

def url(endpoint: Endpoint):
    return f"http://{host}:{port}/api/{endpoint.value}"

def get_state() -> requests.Response:
    mode = requests.get(u := url(Endpoint.status)).json().get('mode')
    logger.debug('%s -> get mode: %s', u, mode)
    return mode

def set_state(state: State) -> requests.Response:
    msg = {"mode": state.value}
    mode = requests.put(u := url(Endpoint.status), json.dumps(msg))
    logger.debug('%s <- set mode: %s', u, state.value)
    return mode

def is_connected() -> bool:

    global exc

    if not utils.is_online(host):
        exc = protocols.TestFailure(f"OpenEphys | No response from {host}: may be offline or unreachable")
        return False

    try:
        state = get_state()
    except requests.RequestException:
        exc = protocols.TestFailure(f"OpenEphys | No response from Open Ephys http server: is the software started?")
        return False
    else:
        if not any(_.value == state for _ in State):
            exc = protocols.TestFailure(f"OpenEphys | Unexpected state: {state}")
            return False

    return True

def initialize() -> None:
    logger.info("OpenEphys | Initializing")
    launch()
    global data_files
    data_files = []
    global initialized
    initialized = time.time()
    set_folder(folder)
    
def get_required_disk_gb() -> int | float:
    """The minimum amount of free disk space required to start recording."""
    return gb_per_hr * min_rec_hr

def is_disk_space_ok() -> bool:
    global exc
    exc = None
    required = get_required_disk_gb()
    for data_root in get_data_roots():
        try:
            free = utils.free_gb(data_root)
        except FileNotFoundError as exc:
            exc = exc
            logger.exception(f"{__name__} data path not accessible: {data_root}")
        else:
            logger.info('%s free disk space on %s: %s GB', __name__, data_root.drive, free)
            if free < required:
                exc = ValueError(f"{__name__} free disk space on {data_root.drive} doesn't meet minimum of {required} GB")
        if exc:
            return False
        return True
       
def test() -> None:
    logger.info("OpenEphys | Testing")
    if not is_connected():
        if exc:
            raise exc
    gb = get_required_disk_gb()
    if not is_disk_space_ok():
        raise TestFailure(f"{__name__} free disk space on one or more recording drives doesn't meet minimum of {gb} GB") from exc

def is_started() -> bool:
    if get_state() == State.record.value:
        return True
    return False

def is_ready_to_start() -> bool:
    if get_state() == State.acquire.value:
        return True
    return False

def start() -> None:
    logger.info('OpenEphys | Starting recording')
    if is_started():
        logger.warning("OpenEphys is already started")
        return
    if not is_ready_to_start():
        set_state(State.acquire)
        time.sleep(.5)
    global latest_start
    latest_start = time.time()
    set_state(State.record)
    
def stop() -> None:
    logger.info('OpenEphys | Stopping recording')
    set_state(State.acquire)

def finalize() -> None:
    logger.info('OpenEphys | Finalizing')
    data_files.extend(get_latest_data_dirs())
    unlock_previous_recording()

def set_folder(name: str,  prepend_text: Optional[str] = "", append_text: Optional[str] = "") -> None:
    """Recording folder string"""
    recording = requests.get(url(Endpoint.recording)).json()
    
    if name == "":
        name = "_" 
        logger.warning("OpenEphys | Recording directory cannot be empty, replaced with underscore: %s" , name)
    if "." in name:
        name.replace(".","_")
        logger.warning("OpenEphys | Recording directory cannot contain periods, replaced with underscores: %s" , name)
    
    recording['base_text'] = name
    recording['prepend_text'] = prepend_text
    recording['append_text'] = append_text
    logger.debug("OpenEphys | Setting recording directory to: %s" , prepend_text+name+append_text)
    response = requests.put(url(Endpoint.recording), json.dumps(recording))
    time.sleep(.1)
    if (actual := response.json().get('base_text')) != name:
        raise protocols.TestFailure(f'OpenEphys | Set folder to {name}, but software shows: {actual}')

def get_folder() -> str | None:
    return requests.get(url(Endpoint.recording)).json().get('base_text')

def clear_open_ephys_name() -> None:
    set_folder("_temp_")

def set_idle():
    "Should be called before sending any configuration to Open Ephys"
    if is_started():
        stop()
    time.sleep(.5)
    set_state(State.idle)
    
def unlock_previous_recording():
    "stop rec/acquiring | set name to _temp_ | record briefly | acquire"
    logger.debug('OpenEphys | Unlocking previous recording')
    set_idle()
    time.sleep(.5)
    clear_open_ephys_name()
    time.sleep(.5)
    start()
    time.sleep(.5)
    stop()
    time.sleep(.5)

def get_record_nodes() -> list[dict[str, Any]]:
    """Returns a list of record node info dicts, incl keys `node_id`, `parent_directory`"""
    return requests.get(url(Endpoint.recording)).json().get('record_nodes', None) or []

def get_data_roots() -> list[pathlib.Path]:
    return [pathlib.Path(f"//{host}/{_['parent_directory'].replace(':','')}") for _ in get_record_nodes()]

def get_latest_data_dirs() -> list[pathlib.Path]:
    """Returns the path to the latest data folder, based on the latest modified time"""
    dirs = []
    for root in get_data_roots():
        if subfolders := [sub for sub in root.iterdir() if sub.is_dir() and not any(_ in str(sub) for _ in ["System Volume Information", "$RECYCLE.BIN"])]:
            subfolders.sort(key=lambda _:_.stat().st_ctime)
            dirs.append(subfolders[-1])
    return dirs
            
def verify() -> None:
    logger.info('OpenEphys | Verifying')
    for data_dir in get_latest_data_dirs():
        for file in reversed(utils.get_files_created_between(data_dir, '*/*/*/continuous/*/sample_numbers.npy', latest_start)):
            if utils.is_file_growing(file):
                break
        else:
            raise protocols.TestFailure(f"OpenEphys | Data file(s) not increasing in size in {data_dir}")
    logger.info('OpenEphys | Verified files are increasing in size for all Record Nodes')
    
def validate() -> None:
    logger.warning('OpenEphys | validate() not implemented')

    
    

  
def set_ref(ext_tip="TIP"):
    # for port in [0, 1, 2]: 
    #     for slot in [0, 1, 2]: 
            
    slot = 2 #! Test
    port = 1 #! Test
    dock = 1 # TODO may be 1 or 2 with firmware upgrade 
    tip_ref_msg = {"text": f"NP REFERENCE {slot} {port} {dock} {ext_tip}"}
    #logger.info(f"sending ...
    # return 
    requests.put('http://localhost:37497/api/processors/100/config', json.dumps(tip_ref_msg))
    time.sleep(3)


    
# TODO set up everything possible from here to avoid accidentally changed settings ? 
# probe channels
# sampling rate 
# tip ref
# signal chain?
# acq drive letters

"""
if __name__ == "__main__":
    r = requests.get(Endpoint.recording)
    print((r.json()['current_directory_name'], r.json()['prepend_text'], r.json()['append_text']))
        
    r = EphysHTTP.set_folder(path = "mouseID_", prepend_text="sessionID", append_text="_date")
    print((r.json()['current_directory_name'], r.json()['prepend_text'], r.json()['append_text']))
    
    r = EphysHTTP.set_folder(path = "mouse", prepend_text="session", append_text="date")
    print((r.json()['current_directory_name'], r.json()['prepend_text'], r.json()['append_text']))
    
    print((r.json()['base_text'])) # fails as of 06/23 https://github.com/open-ephys/plugin-GUI/pull/514
"""


import pdb

# pdb.set_trace()
try:
    import inspect
    import time
    import socket
    import json
    import logging
    import os
    #
    # limstk functions are auto-generated so the ide might warn it can't see them.
    import shutil
    import threading
    from time import sleep

    # import mpetk
    import mpetk.aibsmw.routerio.router as router
    import np.workflows.npxcommon as npxc
    import requests
    import yaml
    from mpetk import limstk, mpeconfig, zro
    # from np.models import model
    from np.models.model import \
        DynamicRouting  # It can make sense to have a class to store experiment data.
    from np.services.ephys_api import EphysHTTP as ephys
    from np.services.mvr import \
        MVRConnector  # This will eventually get incorporated into the workflow launcher
    from wfltk import middleware_messages_pb2 as messages

except Exception as e:
    # import errors aren't printed to console by default
    print(e)

# Setup your typical components; the config, services, and perhaps some typical models for an experiment.
# An alternative is to store these in the state.  However, this is one of those times that globals are ok because
# these values are "read only" and used ubiquitously.

config: dict = mpeconfig.source_configuration("dynamic_routing")

pc = socket.gethostname()
acq = {
    "W10DTSM112719":"W10DTSM112722", # NP0
    "W10DTSM18306":"W10DTSM18278", # NP1
    "W10DTSM18307":"W10DTSM18280", # NP2
    "W10DTMJ0AK6GM":"W10SV108131", #ben-desktop:btTest
}
hostname = acq.get(pc,"localhost")
config['MVR'] = {
    "host": hostname,  
    "port": 50000
}

experiment = DynamicRouting()

io: router.ZMQHandler

mvr: MVRConnector
camstim_agent: zro.Proxy
sync: zro.Proxy
mouse_director: zro.Proxy


def fail_state(message: str, state: dict):
    """
    This is an example of how to create a single failure reporting function.  It inspects the stack to figure out which
    state is reporting the failure and fills out the necessary messaging.  This will post a message to the user and
    allow them to redo the state.

    :param message: the fail reason
    :param state: the state dictionary
    :return:
    """
    current_frame = inspect.currentframe()
    calling_frame = inspect.getouterframes(current_frame, 2)[1][3]
    state_name = calling_frame[:calling_frame.rfind("_")]

    #  These are values expected by the UI
    state['external']['alert'] = True
    state["external"]["transition_result"] = False
    state["external"]["next_state"] = state_name
    state["external"]["msg_text"] = message
    logging.warning(f"{state_name} failed: {message}")


def connect_to_services(state):
    """    The expectation is these services are on (i.e., you have started them with RSC).  If the connection fails you can
    send a message to the user with the fail_state() function above.
    """
    global io
    io = state['resources']['io']

    component_errors = []

    # TODO Here we either need to test open ephys by trying to record or we get the status message.  awaiting testing.

    global mvr
    try:
        mvr = MVRConnector(args=config['MVR'])
        if not mvr._mvr_connected:
            logging.info("Failed to connect to mvr")
            component_errors.append(
                f"Failed to connect to MVR on {config['MVR']}")
    except Exception:
        component_errors.append(f"Failed to connect to MVR.")

    global camstim_agent
    try:
        service = config['camstim_agent']
        camstim_agent = zro.Proxy(
            f"{service['host']}:{service['port']}", timeout=service['timeout'], serialization='json')
        logging.info(f'Camstim Agent Uptime: {camstim_agent.uptime}')
    except Exception:
        component_errors.append(f"Failed to connect to Camstim Agent.")

    global sync
    try:
        service = config['sync']
        sync = zro.Proxy(
            f"{service['host']}:{service['port']}", timeout=service['timeout'])
        logging.info(f'Sync Uptime: {sync.uptime}')
    except Exception:
        component_errors.append(f"Failed to connect to Sync.")

    global mouse_director
    try:
        service = config['mouse_director']
        mouse_director = zro.Proxy(
            f"{service['host']}:{service['port']}", timeout=service['timeout'])
        logging.info(f'MouseDirector Uptime: {mouse_director.uptime}')
    except Exception:
        component_errors.append(f"Failed to connect to MouseDirector.")
        # TODO MDir currently not working - switch this line back on when fixes
        # logging.info(" ** skipping connection to MouseDirector **")

    #  At this point, You could send the user to an informative state to repair the remote services.
    if component_errors:
        fail_state('\n'.join(component_errors), state)


def state_transition(state_transition_function):
    def wrapper(state):

        transition_type = state_transition_function.__name__.split('_')[-1]
        npxc.save_state(state)
        state_transition_function(state)
        return None
    return wrapper


#@state_transition
def resume_enter(state):
    pass


#@state_transition
def resume_input(state):
    prompt = state['external'].get('prompt', False)
    if prompt is True:
        state = npxc.load_previous_state()
    else:
        state['external']['next_state'] = state['external']['next_state']


#@state_transition
def restore_enter(state):
    pass


#@state_transition
def restore_input(state):
    pass


#@state_transition
def init_enter(state):
    """
    Testing image display
    """
    state['external']["exp_logo"] = R"C:\progra~1\AIBS_MPE\workflow_launcher\np\images\logo_np_vis.png"
    connect_to_services(state)

    state["external"]["user_id"] = "ben.hardcastle"
    state["external"]["mouse_id"] = "366122"


#@state_transition
def init_input(state):
    """
    Since this is the first state, you might do some basic startup functionality in the enter state.
    """
    pass


#@state_transition
def get_user_id_entry(state):
    pass


#@state_transition
def get_user_id_input(state):
    """
      Description: The user will input their user name and it will ve validated against the LIMS db.
    """

    # external contains values coming from the UI.  "user_id" is a key specified in the wfl file.
    user_name = state["external"]["user_id"]
    if not limstk.user_details(user_name):
        fail_state(f"Could not find user \"{user_name}\" in LIMS", state)
    state["user_name"] = user_name  # It is ok to save data into the state.
    try:
        mouse_director.set_user_id(user_name)
    except:
        pass


#@state_transition
def get_mouse_id_input(state):
    """
    input function for state get_mouse_id
    :param state: dictionary representing data passing between states and the UI
    """

    # external contains values coming from the UI.  "mouse_id" is a key specified in the wfl file.
    mouse_id = state["external"]["mouse_id"]

    try:
        mouse_result = limstk.donor_info_with_parent(mouse_id)
    except:
        # mouse_result not returned if prev limstk query errors
        fail_state(f"Could not find mouse id \"{mouse_id}\" in LIMS", state)
        return

    #  You might want to get data about this mouse from MTRain
    response = requests.get("http://mtrain:5000/get_script/",
                            data=json.dumps({"LabTracks_ID": mouse_id}))
    if response.status_code != 200:
        fail_state(f"Could not find mouse id \"{mouse_id}\" in MTrain", state)
        return

    try:
        mouse_director.set_mouse_id(mouse_id)
    except:
        pass
    experiment.mouse_id = mouse_id
    experiment.script = response.json()['data']['script'].split('/')[-1]
    experiment.stimulus_name = response.json()['data']['name']

    #  The following is a typical message MPE would expect at this stage.
    log_message = (f"MID, {mouse_id},\n "
                   f'UID, {state["external"]["user_id"].ljust(10)},\n '
                   f'BID, {os.getenv("aibs_comp_id")},\n '
                   f"Action, \n"
                   f"Received\n")
    logging.info(log_message, extra={"weblog": True})



def mvr_capture(state_globals,photo_path="C:/ProgramData/AIBS_MPE/wfltk/temp/last_snapshot.jpg", timeout=30):
    """standard mvr image snapshot func, returning error mesg or img  """
    # pdb.set_trace()         
    mvr.take_snapshot()
    
    def wait_on_snapshot(): 
        t0 = time.time()
        while time.time()-t0 < timeout:
            for message in mvr.read():
                if message.get('mvr_response', "") == "take_snapshot_initiated":
                    source_photo_path = npxc.get_newest_mvr_img(config['MVR']['host'])
                    sleep(1) # MVR has responded too quickly.  It hasn't let go of the file so we must wait.
                    dest_photo_path = shutil.copy2(source_photo_path, photo_path)
                    logging.info(f"Copied: {source_photo_path} -> {dest_photo_path}")
                    return True, dest_photo_path
                elif message.get('mvr_response', "") == "snapshot_failed":
                    return False, message['error_message']  
            # except Exception as e:
            #     return False, e
    success, mesg_or_img = wait_on_snapshot()
    if not success:
        try:
            fail_state(f"Error taking snapshot: {mesg_or_img}",state_globals)
        except: 
            pass
    else:
        return mesg_or_img # return the captured image


#@state_transition
def binary_next_state_prompt(state, tf, next_if_true, next_if_false):
    if tf:
        state["external"]["next_state"] = next_if_true
    else:
        state["external"]["next_state"] = next_if_false


#@state_transition
def pre_brain_surface_photo_doc_enter(state):
    # display new mvr image
    # pdb.set_trace()
    state['external']['new_snapshot'] = npxc.mvr_capture(state)


#@state_transition
def pre_brain_surface_photo_doc_input(state):
    # prompt to retake mvr image
    # pdb.set_trace()
    prompt = state['external'].get('snapshot_retry', False)
    binary_next_state_prompt(state, prompt, next_if_true="pre_brain_surface_photo_doc",
                             next_if_false="probe_insertion_instructions")


#@state_transition
def pre_brain_surface_photo_doc_exit(state):
    #  TODO add to platform json:
    # experiment.platform_json.add_file(dest_photo_path)
    ...


#@state_transition
def flush_lines_enter(state):
    # pdb.set_trace()
    # io["external"][] = "enter"
    # fail_state("testing enter fail", state)
    ...


#@state_transition
def flush_lines_input(state):
    # pdb.set_trace()
    # fail_state("testing input fail", state)
    # io["test2"] = "input"
    ...


#@state_transition
def run_stimulus_enter(state):
    """
    This is a typical method to start recordings and stimulus but is also a model of how to handle user events.
    Note that this is the "enter" function and not the "input" function.  This executes before the screen is presented
    to the user.  The green arrow will not be available to them as a result of the state_busy message.
    """

    #  This is in the enter state because we want to do things before the user sees the screen (like turn off arrow)
    io.write(messages.state_busy(
        message="Waiting for stimulus script to complete."))

    #  Normally, we want to start the recording devices just before camstim
    sync.start()
    mvr.start_record(file_name_prefix=experiment.experiment_id)
    io.write(ephys.start_ecephys_recording())
    #  If you are using the start session api (usually with mtrain, you would do the following.
    camstim_agent.start_session(experiment.mouse_id, experiment.user_name)

    #  TODO:  Add an example of start_script

    #  The following technique for turning off the next arrow and calling a "is ready" function is applicable for
    #  any user event. (timers, waiting on services, etc)

    def check_stimulus():  # you could define this outside of the state of course.
        sleep(5.0)  # gives camstim agent a chance to get started
        while True:
            try:
                if not camstim_agent.is_running():
                    break
                sleep(3.0)
            except Exception as err:
                pass  # time outs are possible depending on the script.  Maybe implement a max retry

        #  It is possible here to check camstim agent for an error and take some action.
        mvr.stop_record()
        io.write(ephys.stop_ecephys_recording())
        sync.stop()
        # re-enables the arrow
        io.write(messages.state_ready(message="Stimulus complete."))
        try:
            mouse_director.finalize_session('')
        except:
            pass
    #  Because the next arrow is disabled, we can wait on this thread to re-enable it without the user being able to
    #  progress the workflow.
    t = threading.Thread(target=check_stimulus)
    t.start()


#@state_transition
def wait_on_sync_enter(state):
    #  This is in the enter state because we want to do things before the user sees the screen (like turn off arrow)
    state["resources"]["io"].write(messages.state_busy(
        message="Waiting for sync to complete."))

    def wait_on_sync():  # you could define this outside of the state of course.
        while "READY" not in sync.get_state():
            sleep(3)
        # re-enables the arrow
        io.write(messages.state_ready(message="Stimulus complete."))

    #  Because the next arrow is disabled, we can wait on this thread to re-enable it without the user being able to
    #  progress the workflow.
    t = threading.Thread(target=wait_on_sync)
    t.start()


#@state_transition
def settle_timer_enter(state):
    #  This is in the enter state because we want to do things before the user sees the screen (like turn off arrow)
    state["resources"]["io"].write(messages.state_busy(
        message="Waiting for stimulus script to complete."))

    def wait_on_timer():  # you could define this outside of the state of course.
        sleep(5.0)  # gives camstim agent a chance to get started
        # re-enables the arrow
        io.write(messages.state_ready(message="Stimulus complete."))

    #  Because the next arrow is disabled, we can wait on this thread to re-enable it without the user being able to
    #  progress the workflow.
    t = threading.Thread(target=wait_on_timer)
    t.start()


def write_manifest():
    lims_session = limstk.Session(
        "neuropixel", "neuropixels", id=experiment.session_id)
    lims_session.trigger_data["sessionid"] = experiment.session_id
    lims_session.trigger_data["location"] = lims_session.path_data["location"]
    # ims_session.add_to_manifest(platform_file_path, remove_source = False)
    # lims_session.write_manifest(trigger_filename=trigger_file_name)



# -*- coding: utf-8 -*-
"""
proxy.py

@author: derricw

Proxy device and manager for ZRO devices.

`Proxy` is a remote object proxy designed to interact with objects
extending `BasePubRepDevice` (from device.py).

"""
import zmq

import np_logging 

logger = np_logging.getLogger(__name__)

def get_address(ip="", port=None):
    """
    Trys to get a properly formatted address given a port and ip.

    ZMQ likes address to be in this format:

        {protocol}://{ip}:{port}

    Args:
        ip (Optional[str]): ip address with some semblance of correct formatting
        port (Optional[int]): port to use in the event that the port is not
            included in the IP

    Returns:
        str: a properly formatted ip str.

    """
    if not ip and not port:
        raise ValueError("Need a port or IP.")
    elif not ip and port:
        return "tcp://*:{}".format(port)
    else:
        if ip[:6] != "tcp://":
            ip = "tcp://"+ip
        if len(ip.split(":")) == 2:
            return "{}:{}".format(ip, port)
        else:
            return ip

class DeviceProxy(object):
    """
    Proxy object for a BasePubRepDevice.

    args:
        ip (Optional[str]): IP/DNS address of actual device. Defaults to
            "localhost".
        port (Optional[int]): REP port for actual device, if it isn't include
            in IP.
        timeout (Optional[float]): Timeout in seconds for all commands.
            Defaults to 10.
        serialization (Optional[str]): Serialization method.  "pickle" (default)
            or "json".

    Example:
        >>> dev = DeviceProxy("localhost:5556")
        >>> dev.call_method_on_device(some_argument)
        >>> dev.attr_on_device = 5

    """
    _context = zmq.Context()
    _context.setsockopt(zmq.LINGER, 1)

    def __init__(self,
                 ip="localhost",
                 port=None,
                 timeout=10.0,
                 serialization='pickle',
                 ):

        super().__init__()
        self.__dict__['ip'] = ip
        self.__dict__['rep_port'] = port
        self.__dict__['timeout'] = timeout
        self.__dict__['serialization'] = serialization.lower()

        self._setup_socket()
        #self._setup_getset()

    def __setattr__(self, name, value):
        """
        Overwrite __setattr__ so that attributes are set on target object
            instead of this object.
        """
        packet = {"command": "set", "args": (name, value)}
        self._send_packet(packet)
        response = self.__dict__['recv']()
        if response == "0":
            return None
        else:
            raise ZroError(message=str(response))

    def __getattr__(self, name):
        """
        Overwrite __getattr__ so that attributes are grabbed from target object
            instead of this object.
        """
        packet = {"command": "get", "args": (name,)}
        self._send_packet(packet)
        response = self.__dict__['recv']()
        if isinstance(response, ZroError):
            raise ZroError(message=str(response))
        elif response in ('callable', "__callable__"):
            self.__dict__['to_call'] = name  # HOLD ON TO YOUR BUTTS
            return self._call
        else:
            return response

    def __dir__(self):
        """
        Overwrite __dir__ so that attributes and methods come from target
            object.
        """
        self.__dict__['to_call'] = "get_attribute_list"
        attrs = self._call()
        self.__dict__['to_call'] = "get_command_list"
        methods = self._call()
        return attrs + methods

    def _send_packet(self, packet):
        """
        Sends a packet.  Attempts to reconnect once if there is a failure.

        #TODO: Make packet a class.
        """
        try:
            self.__dict__['send'](packet)
        except zmq.ZMQError:
            self.__dict__['req_socket'].close()
            self._setup_socket()
            self.__dict__['send'](packet)

    def _setup_socket(self):
        """
        Sets up the request socket.
        """
        ip = self.__dict__['ip']
        rep_port = self.__dict__['rep_port']
        addr_str = get_address(ip, rep_port)
        timeout = self.__dict__['timeout']
        self.__dict__['req_socket'] = self._context.socket(zmq.REQ)
        self.__dict__['req_socket'].setsockopt(zmq.SNDTIMEO, int(timeout*1000))
        self.__dict__['req_socket'].setsockopt(zmq.RCVTIMEO, int(timeout*1000))
        self.__dict__['req_socket'].connect(addr_str)

        if self.__dict__['serialization'] in ["pickle", "pkl", "p"]:
            self.__dict__['send'] = self.__dict__['req_socket'].send_pyobj
            self.__dict__['recv'] = self.__dict__['req_socket'].recv_pyobj
        elif self.__dict__['serialization'] in ["json", "j"]:
            self.__dict__['send'] = self.__dict__['req_socket'].send_json
            self.__dict__['recv'] = self.__dict__['req_socket'].recv_json
        else:
            raise ValueError("Incorrect serialization type. Try 'pickle' or 'json'.")

    def _call(self, *args, **kwargs):
        """
        Used for calling arbitrary methods in the device.
        """
        packet = {"command": "run", "callable": self.to_call, "args": args,
                  "kwargs": kwargs}
        self.__dict__['send'](packet)
        response = self.__dict__['recv']()
        if isinstance(response, dict) and response.get('ZroError', False):
            response = ZroError.from_dict(response)
        if isinstance(response, ZroError):
            raise ZroError(message=str(response))
        return response

    def __del__(self):
        """
        Close the socket on cleanup.
        """
        self.__dict__['req_socket'].close()

Proxy = DeviceProxy

class ZroError(Exception):
    """ Base class for zro errors. """

    error_codes = {
        1: "{} -> HAS_NO_ATTRIBUTE -> {}",
        2: "{} -> HAS_NO_CALLABLE -> {}",
        3: "{} -> ATTRIBUTE_NOT_CALLABLE -> {}",
        4: "{} -> CALLABLE_FAILED -> {}",
        5: "{} -> ARGUMENTS_INVALID -> {}",
        6: "{} -> UNHANDLED_ERROR -> {}",
        7: "{} -> ASYNC_RESULT_INVALID_HANDLE -> {}",
        8: "{} -> ASYNC_RESULT_UNFINISHED -> {}",
        9: "{} -> ASYNC_CALLBACK_FAILED -> {}",
    }

    HAS_NO_ATTRIBUTE = 1
    HAS_NO_CALLABLE = 2
    ATTRIBUTE_NOT_CALLABLE = 3
    CALLABLE_FAILED = 4
    ARGUMENTS_INVALID = 5
    UNHANDLED_ERROR = 6
    ASYNC_RESULT_INVALID_HANDLE = 7
    ASYNC_RESULT_UNFINISHED = 8
    ASYNC_CALLBACK_FAILED = 9

    def __init__(self, obj=None, target=None, error_code=6, message=""):
        if not message:
            message = self.error_codes[error_code].format(obj, target)
        self.message = message
        self.error_code = error_code
        super(ZroError, self).__init__(message)

    def to_JSON(self):
        return {
            'ZroError': str(type(self.get_specific_error())), # this key lets zro convert this on the receive side
            'error_code': self.error_code,
            'message': str(self.message)
        }

    @staticmethod
    def from_dict(d):
        return ZroError(error_code=d['error_code'], message=d['message']).get_specific_error()

    def get_specific_error(self, to_raise=False):
        """ Get the appropriate ZroError for the error type. """
        err = _SPECIFIC_ERRORS[self.error_code](message=self.message)
        if to_raise:
            raise err
        return err


class ZroNoAttributeError(ZroError):
    """ Error for HAS_NO_ATTRIBUTE. """
    def __init__(self, obj=None, target=None, message=""):
        super(ZroNoAttributeError, self).__init__(
            obj, target, ZroError.HAS_NO_ATTRIBUTE, message)


class ZroNoCallableError(ZroError):
    """ Error for HAS_NO_CALLABLE. """
    def __init__(self, obj=None, target=None, message=""):
        super(ZroNoCallableError, self).__init__(
            obj, target, ZroError.HAS_NO_CALLABLE, message)


class ZroAttrNotCallableError(ZroError):
    """ Error for ATTRIBUTE_NOT_CALLABLE. """
    def __init__(self, obj=None, target=None, message=""):
        super(ZroAttrNotCallableError, self).__init__(
            obj, target, ZroError.ATTRIBUTE_NOT_CALLABLE, message)


class ZroCallableFailedError(ZroError):
    """ Error for CALLABLE_FAILED. """
    def __init__(self, obj=None, target=None, message=""):
        super(ZroCallableFailedError, self).__init__(
            obj, target, ZroError.CALLABLE_FAILED, message)


class ZroArgumentsInvalidError(ZroError):
    """ Error for ARGUMENTS_INVALID. """
    def __init__(self, obj=None, target=None, message=""):
        super(ZroArgumentsInvalidError, self).__init__(
            obj, target, ZroError.ARGUMENTS_INVALID, message)


class ZroAsyncHandleInvalidError(ZroError):
    """ Error for ASYNC_RESULT_INVALID_HANDLE. """
    def __init__(self, obj=None, target=None, message=""):
        super(ZroAsyncHandleInvalidError, self).__init__(
            obj, target, ZroError.ASYNC_RESULT_INVALID_HANDLE, message)


class ZroResultUnfinishedError(ZroError):
    """ Error for ASYNC_RESULT_UNFINISHED. """
    def __init__(self, obj=None, target=None, message=""):
        super(ZroResultUnfinishedError, self).__init__(
            obj, target, ZroError.ASYNC_RESULT_UNFINISHED, message)


class ZroCallbackFailedError(ZroError):
    """ Error for ASYNC_CALLBACK_FAILED. """
    def __init__(self, obj=None, target=None, message=""):
        super(ZroCallbackFailedError, self).__init__(
            obj, target, ZroError.ASYNC_CALLBACK_FAILED, message)


_SPECIFIC_ERRORS = {
    1: ZroNoAttributeError,
    2: ZroNoCallableError,
    3: ZroAttrNotCallableError,
    4: ZroCallableFailedError,
    5: ZroArgumentsInvalidError,
    6: ZroError,
    7: ZroAsyncHandleInvalidError,
    8: ZroResultUnfinishedError,
    9: ZroCallbackFailedError,
}



from __future__ import annotations

import datetime
import pathlib
import typing
from typing import Any, Optional, Protocol, Union, Sequence, ClassVar, Type

import np_config
import np_session

from .baseclasses import Experiment
from ..services.protocols import Service

RESUME_ROOT = pathlib.Path(np_config.from_zk('/projects/np_workflows/defaults/configuration')['workflow_resume_root'])
RESUME_DATETIME_FMT = '%Y-%m-%d_%H%M%S'

def save(experiment: Experiment, suffix: Optional[str] = None):
    path = RESUME_ROOT / experiment.__class__.__name__ / experiment.session 
    filename = datetime.datetime.now().strftime(RESUME_DATETIME_FMT)
    if suffix:
        filename = f'{filename}_{suffix}'
    path.mkdir(exist_ok=True, parents=True)
    pickle.dump(experiment, path / f'{filename}.pkl')

def sorted_dirs(path: pathlib.Path) -> list[pathlib.Path]:
    return sorted([p for p in path.iterdir() if p.is_dir()], key=lambda _:_.stat().ctime, reverse=True)

def most_recent_session(experiment_type: Type[Experiment]) -> pathlib.Path | None:
    root = RESUME_ROOT / experiment_type.__name__
    if root.exists() and (dirs := sorted_dirs(root)):
        return dirs[0]
    return None 
        
def saves_from_most_recent_session(experiment_type: Type[Experiment]) -> list[pathlib.Path] | None:
    if session_dir := most_recent_session(experiment_type):
        return sorted_dirs(session_dir)
    return None 

def most_recent_save(experiment_type: Type[Experiment]) -> pathlib.Path | None:
    "Path to most recent save state"
    if previous := saves_from_most_recent_session(experiment_type):
        return sorted(previous.glob('*.pkl'), key=lambda _:_.name)
    return None

def load(path: pathlib.Path):
    return pickle.load(path)

        